# - coding: utf-8 -
import struct
from collections import namedtuple

from .base import Request, Response, ConsumerGroupProtocolMetadata, MemberAssignment
from ..utils import struct_helpers


class ListGroupsRequest(Request):
    """A list groups request

    Specification::

    ListGroupsRequest =>
    """
    API_KEY = 16

    def get_bytes(self):
        """Create a new list group request"""
        output = bytearray(len(self))
        self._write_header(output)
        return output

    def __len__(self):
        """Length of the serialized message, in bytes"""
        return self.HEADER_LEN


GroupListing = namedtuple(
    'GroupListing',
    ['group_id', 'protocol_type']
)


class ListGroupsResponse(Response):
    """A list groups response

    Specification::

    ListGroupsResponse => ErrorCode Groups
      ErrorCode => int16
      Groups => [GroupId ProtocolType]
        GroupId => string
        ProtocolType => string
    """
    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'h [SS]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.error = response[0]
        self.groups = {}
        for group_info in response[1]:
            listing = GroupListing(*group_info)
            self.groups[listing.group_id] = listing


class DescribeGroupsRequest(Request):
    """A describe groups request

    Specification::

    DescribeGroupsRequest => [GroupId]
      GroupId => string
    """
    API_KEY = 15

    def __init__(self, group_ids):
        self.group_ids = group_ids

    def get_bytes(self):
        """Create a new list group request"""
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!i'
        struct.pack_into(fmt, output, offset, len(self.group_ids))
        offset += struct.calcsize(fmt)
        for group_id in self.group_ids:
            fmt = '!h%ds' % len(group_id)
            struct.pack_into(fmt, output, offset, len(group_id), group_id)
            offset += struct.calcsize(fmt)
        return output

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # header + len(group_ids)
        size = self.HEADER_LEN + 4
        for group_id in self.group_ids:
            # len(group_id) + group_id
            size += 2 + len(group_id)
        return size


GroupMember = namedtuple(
    'GroupMember',
    ['member_id', 'client_id', 'client_host', 'member_metadata', 'member_assignment']
)


DescribeGroupResponse = namedtuple(
    'DescribeGroupResponse',
    ['error_code', 'group_id', 'state', 'protocol_type', 'protocol', 'members']
)


class DescribeGroupsResponse(Response):
    """A describe groups response

    Specification::


    DescribeGroupsResponse => [ErrorCode GroupId State ProtocolType Protocol Members]
      ErrorCode => int16
      GroupId => string
      State => string
      ProtocolType => string
      Protocol => string
      Members => [MemberId ClientId ClientHost MemberMetadata MemberAssignment]
        MemberId => string
        ClientId => string
        ClientHost => string
        MemberMetadata => bytes
        MemberAssignment => bytes
    """
    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[hSSSS [SSSYY ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.groups = {}
        for group_info in response:
            members = {}
            for member_info in group_info[5]:
                member_metadata = ConsumerGroupProtocolMetadata.from_bytestring(
                    member_info[3])
                member_assignment = MemberAssignment.from_bytestring(member_info[4])
                member = GroupMember(*(member_info[:3] + (member_metadata,
                                                          member_assignment)))
                members[member.member_id] = member
            group = DescribeGroupResponse(*(group_info[:5] + (members,)))
            self.groups[group.group_id] = group


_CreateTopicRequest = namedtuple(
    'CreateTopicRequest',
    ['topic_name', 'num_partitions', 'replication_factor', 'replica_assignment',
     'config_entries']
)


class CreateTopicRequest(_CreateTopicRequest):
    def __new__(cls,
                topic_name,
                num_partitions,
                replication_factor,
                replica_assignment,
                config_entries):
        return super(CreateTopicRequest, cls).__new__(
            cls, topic_name, num_partitions, replication_factor, replica_assignment,
            config_entries)


class CreateTopicsRequest(Request):
    """A create topics request

    Specification::

    CreateTopics Request (Version: 0) => [create_topic_requests] timeout
        create_topic_requests => topic num_partitions replication_factor [replica_assignment] [config_entries]
            topic => STRING
            num_partitions => INT32
            replication_factor => INT16
            replica_assignment => partition [replicas]
                partition => INT32
                replicas => INT32
            config_entries => config_name config_value
                config_name => STRING
                config_value => NULLABLE_STRING
        timeout => INT32
    """
    API_KEY = 19

    def __init__(self, topic_requests, timeout=0):
        self.topic_requests = topic_requests
        self.timeout = timeout

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # header + len(topic_reqs)
        size = self.HEADER_LEN + 4
        for topic_req in self.topic_requests:
            # len(topic_name) + topic_name
            size += 2 + len(topic_req.topic_name)
            # num_partitions + replication_factor + len(replica_assignment)
            size += 4 + 2 + 4
            for partition, replicas in topic_req.replica_assignment:
                # partition + len(replicas) + replicas
                size += 4 + 4 + 4 * len(replicas)
            # len(config_entries)
            size += 4
            for config_name, config_value in topic_req.config_entries:
                # len(config_name) + config_name + len(config_value) + config_value
                size += 2 + len(config_name) + 2 + len(config_value)
        # timeout
        size += 4
        return size

    def get_bytes(self):
        """Create a new create topics request"""
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!i'
        struct.pack_into(fmt, output, offset, len(self.topic_requests))
        offset += struct.calcsize(fmt)
        for topic_req in self.topic_requests:
            fmt = '!h%dsihi' % len(topic_req.topic_name)
            struct.pack_into(fmt, output, offset, len(topic_req.topic_name),
                             topic_req.topic_name, topic_req.num_partitions,
                             topic_req.replication_factor,
                             len(topic_req.replica_assignment))
            offset += struct.calcsize(fmt)
            for partition, replicas in topic_req.replica_assignment:
                fmt = '!ii'
                struct.pack_into(fmt, output, offset, partition, len(replicas))
                offset += struct.calcsize(fmt)
                for replica in replicas:
                    fmt = '!i'
                    struct.pack_into(fmt, output, offset, replica)
                    offset += struct.calcsize(fmt)
            fmt = '!i'
            struct.pack_into(fmt, output, offset, len(topic_req.config_entries))
            offset += struct.calcsize(fmt)
            for config_name, config_value in topic_req.config_entries:
                fmt = '!h%dsh%ds' % (len(config_name), len(config_value))
                struct.pack_into(fmt, output, offset, len(config_name), config_name,
                                 len(config_value), config_value)
                offset += struct.calcsize(fmt)
        fmt = '!i'
        struct.pack_into(fmt, output, offset, self.timeout)
        offset += struct.calcsize(fmt)
        return output


class CreateTopicsResponse(Response):
    """A create topics response

    Specification::

    CreateTopics Response (Version: 0) => [topic_errors]
        topic_errors => topic error_code
            topic => STRING
            error_code => INT16
    """
    API_KEY = 19

    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[Sh]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        for _, error_code in response:
            if error_code != 0:
                self.raise_error(error_code, response)


class DeleteTopicsRequest(Request):
    """A delete topics request

    Specification::

    DeleteTopics Request (Version: 0) => [topics] timeout
        topics => STRING
        timeout => INT32
    """
    API_KEY = 20

    def __init__(self, topics, timeout=0):
        self.topics = topics
        self.timeout = timeout

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # header + len(topics)
        size = self.HEADER_LEN + 4
        for topic in self.topics:
            # len(topic) + group_id
            size += 2 + len(topic)
        # timeout
        size += 4
        return size

    def get_bytes(self):
        """Create a new delete topics request"""
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!i'
        struct.pack_into(fmt, output, offset, len(self.topics))
        offset += struct.calcsize(fmt)
        for topic in self.topics:
            fmt = '!h%ds' % len(topic)
            struct.pack_into(fmt, output, offset, len(topic), topic)
            offset += struct.calcsize(fmt)
        fmt = '!i'
        struct.pack_into(fmt, output, offset, self.timeout)
        offset += struct.calcsize(fmt)
        return output


class DeleteTopicsResponse(Response):
    """A delete topics response

    Specification::

    DeleteTopics Response (Version: 0) => [topic_error_codes]
        topic_error_codes => topic error_code
            topic => STRING
            error_code => INT16
    """
    API_KEY = 20

    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[Sh]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        for _, error_code in response:
            if error_code != 0:
                self.raise_error(error_code, response)


class ApiVersionsRequest(Request):
    """An api versions request

    Specification::

        ApiVersions Request (Version: 0) =>
    """
    API_KEY = 18

    def get_bytes(self):
        """Create a new api versions request"""
        output = bytearray(len(self))
        self._write_header(output)
        return output

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # header
        size = self.HEADER_LEN
        return size


ApiVersionsSpec = namedtuple('ApiVersionsSpec', ['key', 'min', 'max'])


class ApiVersionsResponse(Response):
    """
    Specification::

    ApiVersions Response (Version: 0) => error_code [api_versions]
        error_code => INT16
        api_versions => api_key min_version max_version
            api_key => INT16
            min_version => INT16
            max_version => INT16
    """
    API_KEY = 18

    @classmethod
    def get_versions(cls):
        return {0: ApiVersionsResponse, 1: ApiVersionsResponseV1}

    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'h [hhh]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.api_versions = {}
        for api_key, min_v, max_v in response[1]:
            self.api_versions[api_key] = ApiVersionsSpec(api_key, min_v, max_v)


class ApiVersionsResponseV1(ApiVersionsResponse):
    """
    Specification::

    ApiVersions Response (Version: 1) => error_code [api_versions] throttle_time_ms
        error_code => INT16
        api_versions => api_key min_version max_version
            api_key => INT16
            min_version => INT16
            max_version => INT16
        throttle_time_ms => INT32
    """
    def __init__(self, buff):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'h [hhh]i'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.api_versions = {}
        for api_key, min_v, max_v in response[1]:
            self.api_versions[api_key] = ApiVersionsSpec(api_key, min_v, max_v)
        self.throttle_time = response[2]



import pytest

from tests.pykafka import test_simpleconsumer, test_balancedconsumer
from pykafka.utils.compat import range
try:
    from pykafka.rdkafka import _rd_kafka  # noqa
    RDKAFKA = True
except ImportError:
    RDKAFKA = False  # C extension not built


@pytest.mark.skipif(not RDKAFKA, reason="rdkafka")
class TestRdKafkaSimpleConsumer(test_simpleconsumer.TestSimpleConsumer):
    USE_RDKAFKA = True

    def test_update_cluster(self):
        """Won't work because we don't run SimpleConsumer.fetch"""
        super(TestRdKafkaSimpleConsumer, self).test_update_cluster()

    def test_offset_commit_agrees(self):
        """Check rdkafka-obtained offsets arrive correctly

        In RdKafkaSimpleConsumer.consume we bypass most of the internals of
        simpleconsumer.OwnedPartition, but then expect it to still commit
        offsets for us correctly.  This warrants very explicit testing.
        """
        with self._get_simple_consumer(
                consumer_group=b'test_offset_commit_agrees') as consumer:
            latest_offs = _latest_partition_offsets_by_reading(consumer, 100)
            consumer.commit_offsets()

            # We can only compare partitions we've consumed from, so filter:
            retrieved_offs = {r[0]: r[1].offset - 1
                              for r in consumer.fetch_offsets()
                              if r[0] in latest_offs}
            self.assertEquals(retrieved_offs, latest_offs)

    def test_offset_resume_agrees(self):
        """Check the rdkafka consumer returns messages at specified offset

        Make sure reads from the underlying rdkafka consumer really do start
        at the offsets dictated by SimpleConsumer
        """
        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume_agrees') as consumer:
            latest_offs = _latest_partition_offsets_by_reading(consumer, 100)
            consumer.commit_offsets()

        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume_agrees') as consumer:
            # check each partition, then tick it off:
            while latest_offs:
                msg = consumer.consume()
                if msg.partition_id not in latest_offs:
                    # ie we didn't get to this partition previously
                    continue
                expected_offset = latest_offs[msg.partition_id] + 1
                self.assertEquals(msg.offset, expected_offset)
                del latest_offs[msg.partition_id]


def _latest_partition_offsets_by_reading(consumer, n_reads):
    """Obtain message offsets from consumer, return grouped by partition"""
    latest_offs = {}
    for _ in range(n_reads):
        msg = consumer.consume()
        latest_offs[msg.partition_id] = msg.offset
    return latest_offs


@pytest.mark.skipif(not RDKAFKA, reason="rdkafka")
class RdkBalancedConsumerIntegrationTests(test_balancedconsumer.BalancedConsumerIntegrationTests):
    USE_RDKAFKA = True



from __future__ import division

import mock
import os
import platform
import pytest
import random
import time
import types
import unittest2
from uuid import uuid4

try:
    import gevent
except ImportError:
    gevent = None

try:
    from pykafka.rdkafka import _rd_kafka  # noqa
    RDKAFKA = True
except ImportError:
    RDKAFKA = False  # C extension not built

from pykafka import KafkaClient
from pykafka.common import OffsetType
from pykafka.exceptions import (MessageSizeTooLarge, ProducerQueueFullError,
                                ProduceFailureError)
from pykafka.partitioners import hashing_partitioner
from pykafka.protocol import Message
from pykafka.test.utils import get_cluster, stop_cluster, retry
from pykafka.common import CompressionType
from pykafka.producer import OwnedBroker
from pykafka.utils import serialize_utf8, deserialize_utf8

kafka_version = os.environ.get('KAFKA_VERSION', '0.8.0')


class ProducerIntegrationTests(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = b'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.client = KafkaClient(cls.kafka.brokers,
                                 use_greenlets=cls.USE_GEVENT,
                                 broker_version=kafka_version)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def _get_producer(self, **kwargs):
        topic = self.client.topics[self.topic_name]
        return topic.get_producer(use_rdkafka=self.USE_RDKAFKA, **kwargs)

    def _get_consumer(self, **kwargs):
        return self.client.topics[self.topic_name].get_simple_consumer(
            consumer_timeout_ms=1000,
            auto_offset_reset=OffsetType.LATEST,
            reset_offset_on_start=True,
            **kwargs
        )

    def test_produce(self):
        # unique bytes, just to be absolutely sure we're not fetching data
        # produced in a previous test
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(sync=True, min_queued_messages=1)
        prod.produce(payload)

        message = consumer.consume()
        assert message.value == payload

    def test_produce_utf8(self):
        payload = u"{}".format(random.random())
        consumer = self._get_consumer(deserializer=deserialize_utf8)
        prod = self._get_producer(sync=True, min_queued_messages=1,
                                  serializer=serialize_utf8)
        prod.produce(payload)
        message = consumer.consume()
        assert message.value == payload

    def test_sync_produce_raises(self):
        """Ensure response errors are raised in produce() if sync=True"""
        with self._get_producer(sync=True, min_queued_messages=1) as prod:
            with self.assertRaises(MessageSizeTooLarge):
                prod.produce(10 ** 7 * b" ")

        if not self.USE_RDKAFKA:
            # ensure that a crash on a worker thread still raises exception in sync mode
            p = self._get_producer(sync=True)

            def stub_send_request(self, message_batch, owned_broker):
                1 / 0
            p._send_request = types.MethodType(stub_send_request, p)
            with self.assertRaises(ZeroDivisionError):
                p.produce(b"test")

    def test_sync_produce_doesnt_hang(self):
        producer = self._get_producer(sync=True)

        def stub_mark(w, x, y, z):
            return None
        # simulate delivery report being lost
        producer._mark_as_delivered = types.MethodType(stub_mark, producer)
        producer._delivery_reports = mock.MagicMock()
        with self.assertRaises(ProduceFailureError):
            producer.produce(b"test")

    def test_produce_hashing_partitioner(self):
        # unique bytes, just to be absolutely sure we're not fetching data
        # produced in a previous test
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(
            sync=True,
            min_queued_messages=1,
            partitioner=hashing_partitioner)
        prod.produce(payload, partition_key=b"dummy")

        # set a timeout so we don't wait forever if we break producer code
        message = consumer.consume()
        assert message.value == payload

    def test_async_produce(self):
        payload = uuid4().bytes

        consumer = self._get_consumer()

        prod = self._get_producer(min_queued_messages=1, delivery_reports=True)
        prod.produce(payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, payload)
        self.assertIsNone(report[1])
        self.assertGreaterEqual(report[0].offset, 0)

        message = consumer.consume()
        assert message.value == payload

    def test_recover_disconnected(self):
        """Test our retry-loop with a recoverable error"""
        payload = uuid4().bytes
        prod = self._get_producer(min_queued_messages=1, delivery_reports=True)

        for broker in self.client.brokers.values():
            broker._connection.disconnect()

        prod.produce(payload)
        report = prod.get_delivery_report()
        self.assertIsNone(report[1])

    def test_async_produce_context(self):
        """Ensure that the producer works as a context manager"""
        payload = uuid4().bytes

        consumer = self._get_consumer()
        with self._get_producer(min_queued_messages=1) as producer:
            producer.produce(payload)

        message = consumer.consume()
        assert message.value == payload

    def test_async_produce_queue_full(self):
        """Ensure that the producer raises an error when its queue is full"""
        consumer = self._get_consumer()
        with self._get_producer(block_on_queue_full=False,
                                max_queued_messages=1,
                                linger_ms=1000) as producer:
            with self.assertRaises(ProducerQueueFullError):
                while True:
                    producer.produce(uuid4().bytes)
        while consumer.consume() is not None:
            time.sleep(.05)

    def test_async_produce_lingers(self):
        """Ensure that the context manager waits for linger_ms milliseconds"""
        if self.USE_RDKAFKA:
            pytest.skip("rdkafka uses different lingering mechanism")
        linger = 3
        consumer = self._get_consumer()
        with self._get_producer(linger_ms=linger * 1000) as producer:
            start = time.time()
            producer.produce(uuid4().bytes)
            producer.produce(uuid4().bytes)
        self.assertTrue(int(time.time() - start) >= int(linger))
        consumer.consume()
        consumer.consume()

    def test_async_produce_thread_exception(self):
        """Ensure that an exception on a worker thread is raised to the main thread"""
        consumer = self._get_consumer()
        with self.assertRaises(AttributeError):
            with self._get_producer(min_queued_messages=1) as producer:
                # get some dummy data into the queue that will cause a crash
                # when flushed:
                msg = Message("stuff", partition_id=0)
                del msg.value
                producer._produce(msg)
        while consumer.consume() is not None:
            time.sleep(.05)

    def test_required_acks(self):
        """Test with non-default values for `required_acks`

        See #278 for a related bug.  Here, we only test that no exceptions
        occur (hence `sync=True`, which would surface most exceptions)
        """
        kwargs = dict(linger_ms=1, sync=True, required_acks=0)
        prod = self._get_producer(**kwargs)
        prod.produce(uuid4().bytes)

        kwargs["required_acks"] = -1
        prod = self._get_producer(**kwargs)
        prod.produce(uuid4().bytes)

    def test_null_payloads(self):
        """Test that None is accepted as a null payload"""
        consumer = self._get_consumer()
        prod = self._get_producer(sync=True, min_queued_messages=1)
        prod.produce(None)
        self.assertIsNone(consumer.consume().value)
        prod.produce(None, partition_key=b"whatever")
        self.assertIsNone(consumer.consume().value)
        prod.produce(b"")  # empty string should be distinguished from None
        self.assertEqual(b"", consumer.consume().value)

    def test_owned_broker_flush_message_larger_then_max_request_size(self):
        """Test that producer batches messages into the batches no larger than
        `max_request_size`
        """
        large_payload = b''.join([uuid4().bytes for i in range(50000)])

        producer = self._get_producer(
            auto_start=False
        )
        # setup producer but do not start actually sending messages
        partition = producer._topic.partitions[0]
        owned_broker = OwnedBroker(producer, partition.leader, auto_start=False)

        delivery_report_queue = producer._cluster.handler.Queue()
        msg = Message(
            large_payload,
            partition_id=0,
            delivery_report_q=delivery_report_queue
        )

        owned_broker.enqueue(msg)

        max_request_size = 1000
        assert max_request_size < len(msg)
        owned_broker.flush(0, max_request_size)
        q_msg, exc = delivery_report_queue.get()
        assert q_msg is msg
        assert isinstance(exc, MessageSizeTooLarge)

    def test_owned_broker_flush_batching_by_max_request_size(self):
        """Test that producer batches messages into the batches no larger than
        `max_request_size`
        """
        large_payload = b''.join([uuid4().bytes for i in range(5000)])

        producer = self._get_producer(
            auto_start=False
        )
        # setup producer but do not start actually sending messages
        partition = producer._topic.partitions[0]
        owned_broker = OwnedBroker(producer, partition.leader, auto_start=False)

        for i in range(100):
            msg = Message(large_payload, partition_id=0)
            owned_broker.enqueue(msg)

        batch = owned_broker.flush(0, producer._max_request_size)
        assert len(batch) < 100
        assert sum([len(m.value) for m in batch]) < producer._max_request_size

        # iterate through the rest of the batches and test the same invariant
        while batch:
            batch = owned_broker.flush(0, producer._max_request_size)
            assert len(batch) < 100
            assert sum([len(m.value) for m in batch]) < producer._max_request_size

    def test_async_produce_compression_large_message(self):
        # TODO: make payload size bigger once pypy snappy compression issue is
        # fixed
        large_payload = b''.join([uuid4().bytes for i in range(5)])
        consumer = self._get_consumer()

        prod = self._get_producer(
            compression=CompressionType.SNAPPY,
            delivery_reports=True
        )
        prod.produce(large_payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, large_payload)
        self.assertIsNone(report[1])

        message = consumer.consume()
        assert message.value == large_payload

        for i in range(10):
            prod.produce(large_payload)

        # use retry logic to loop over delivery reports and ensure we can
        # produce a group of large messages
        reports = []

        def ensure_all_messages_produced():
            report = prod.get_delivery_report()
            reports.append(report)
            assert len(reports) == 10
        retry(ensure_all_messages_produced, retry_time=30, wait_between_tries=0.5)

        for report in reports:
            self.assertEqual(report[0].value, large_payload)
            self.assertIsNone(report[1])

        # cleanup and consumer all messages
        msgs = []

        def ensure_all_messages_consumed():
            msg = consumer.consume()
            if msg:
                msgs.append(msg)
            assert len(msgs) == 10
        retry(ensure_all_messages_consumed, retry_time=15)

    def test_async_produce_large_message(self):

        consumer = self._get_consumer()
        large_payload = b''.join([uuid4().bytes for i in range(50000)])
        assert len(large_payload) / 1024 / 1024 < 1.0

        prod = self._get_producer(delivery_reports=True, linger_ms=1000)
        prod.produce(large_payload)

        report = prod.get_delivery_report()
        self.assertEqual(report[0].value, large_payload)
        self.assertIsNone(report[1])

        message = consumer.consume()
        assert message.value == large_payload

        for i in range(10):
            prod.produce(large_payload)

        # use retry logic to loop over delivery reports and ensure we can
        # produce a group of large messages
        reports = []

        def ensure_all_messages_produced():
            report = prod.get_delivery_report()
            reports.append(report)
            assert len(reports) == 10
        retry(ensure_all_messages_produced, retry_time=30, wait_between_tries=0.5)

        for report in reports:
            self.assertEqual(report[0].value, large_payload)
            self.assertIsNone(report[1])

        # cleanup and consumer all messages
        msgs = []

        def ensure_all_messages_consumed():
            msg = consumer.consume()
            if msg:
                msgs.append(msg)
            assert len(msgs) == 10
        retry(ensure_all_messages_consumed, retry_time=15)


@pytest.mark.skipif(not RDKAFKA, reason="rdkafka")
class TestRdKafkaProducer(ProducerIntegrationTests):
    USE_RDKAFKA = True


@pytest.mark.skipif(platform.python_implementation() == "PyPy" or gevent is None,
                    reason="Unresolved crashes")
class TestGEventProducer(ProducerIntegrationTests):
    USE_GEVENT = True


if __name__ == "__main__":
    unittest2.main()



import os
import unittest
from uuid import uuid4

import pytest

from pykafka import KafkaClient, SslConfig
from pykafka.test.utils import get_cluster, stop_cluster


kafka_version = os.environ.get('KAFKA_VERSION', '0.8.0')


class SslIntegrationTests(unittest.TestCase):
    USE_RDKAFKA = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        if cls.kafka.brokers_ssl is None:
            pytest.skip("Test-cluster doesn't advertise ssl ports.")

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def roundtrip_test(self, client):
        """Test producing then consuming

        This is mostly important to test the pykafka.rdkafka classes, which
        should be passed SSL settings during producer/consumer init.
        """
        topic_name = uuid4().hex.encode()
        payload = uuid4().hex.encode()
        topic = client.topics[topic_name]

        producer = topic.get_producer(use_rdkafka=self.USE_RDKAFKA, sync=True)
        producer.produce(payload)

        consumer = topic.get_simple_consumer(use_rdkafka=self.USE_RDKAFKA,
                                             consumer_timeout_ms=5000)
        self.assertEqual(consumer.consume().value, payload)

    def test_ca_only(self):
        """Connect with CA cert only (ie no client cert)"""
        config = SslConfig(cafile=self.kafka.certs.root_cert)
        client = KafkaClient(self.kafka.brokers_ssl, ssl_config=config,
                             broker_version=kafka_version)
        self.roundtrip_test(client)

    def test_client_cert(self):
        """Connect with client certificate"""
        # This would be a more potent test if we could on-the-fly reconfigure
        # the test cluster to refuse connections without client certs, but
        # that's not easy to achieve with our current setup
        certs = self.kafka.certs
        config = SslConfig(cafile=certs.root_cert,
                           certfile=certs.client_cert,
                           keyfile=certs.client_key,
                           password=certs.client_pass)
        client = KafkaClient(self.kafka.brokers_ssl, ssl_config=config,
                             broker_version=kafka_version)
        self.roundtrip_test(client)

    @pytest.mark.skip(reason="Unresolved crashes")
    def test_legacy_wrap_socket(self):
        """Test socket-wrapping without SSLContext"""
        config = SslConfig(cafile=self.kafka.certs.root_cert)
        config._wrap_socket = config._legacy_wrap_socket()
        client = KafkaClient(self.kafka.brokers_ssl, ssl_config=config,
                             broker_version=kafka_version)
        self.roundtrip_test(client)



# - coding: utf-8 -
import struct
from collections import namedtuple, defaultdict

from .base import Request, Response
from ..utils import struct_helpers
from ..utils.compat import iteritems


_PartitionOffsetRequest = namedtuple(
    'PartitionOffsetRequest',
    ['topic_name', 'partition_id', 'offsets_before', 'max_offsets']
)


class PartitionOffsetRequest(_PartitionOffsetRequest):
    """Offset request for a specific topic/partition
    :ivar topic_name: Name of the topic to look up
    :ivar partition_id: Id of the partition to look up
    :ivar offsets_before: Retrieve offset information for messages before
                          this timestamp (ms). -1 will retrieve the latest
                          offsets and -2 will retrieve the earliest
                          available offset. If -2,only 1 offset is returned
    :ivar max_offsets: How many offsets to return
    """
    pass


class ListOffsetRequest(Request):
    """An offset request
    Specification::
        ListOffsetRequest => ReplicaId [TopicName [Partition Time MaxNumberOfOffsets]]
          ReplicaId => int32
          TopicName => string
          Partition => int32
          Time => int64
          MaxNumberOfOffsets => int32
    """
    API_VERSION = 0
    API_KEY = 2

    @classmethod
    def get_versions(cls):
        # XXX use ListOffsetRequestV1 after 0.10 message format is supported
        return {0: ListOffsetRequest, 1: ListOffsetRequest}

    def __init__(self, partition_requests):
        """Create a new offset request"""
        self._reqs = defaultdict(dict)
        for t in partition_requests:
            self._reqs[t.topic_name][t.partition_id] = (t.offsets_before,
                                                        t.max_offsets)

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + replicaId + len(topics)
        size = self.HEADER_LEN + 4 + 4
        for topic, parts in iteritems(self._reqs):
            # topic name + len(parts)
            size += 2 + len(topic) + 4
            # partition + fetch offset + max bytes => for each partition
            size += (4 + 8 + 4) * len(parts)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output, api_version=self.API_VERSION)
        offset = self.HEADER_LEN
        struct.pack_into('!ii', output, offset, -1, len(self._reqs))
        offset += 8
        for topic_name, partitions in iteritems(self._reqs):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name),
                             topic_name, len(partitions))
            offset += struct.calcsize(fmt)
            for pnum, (offsets_before, max_offsets) in iteritems(partitions):
                struct.pack_into('!iqi', output, offset,
                                 pnum, offsets_before, max_offsets)
                offset += 16
        return output


class ListOffsetRequestV1(ListOffsetRequest):
    """
    Specification::
        ListOffsetRequest => ReplicaId [TopicName [Partition Time]]
          ReplicaId => int32
          TopicName => string
          Partition => int32
          Time => int64
    """
    API_VERSION = 1

    def __init__(self, partition_requests):
        """Create a new offset request"""
        self._reqs = defaultdict(dict)
        for t in partition_requests:
            self._reqs[t.topic_name][t.partition_id] = t.offsets_before

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + replicaId + len(topics)
        size = self.HEADER_LEN + 4 + 4
        for topic, parts in iteritems(self._reqs):
            # topic name + len(parts)
            size += 2 + len(topic) + 4
            # partition + fetch offset => for each partition
            size += (4 + 8) * len(parts)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output, api_version=self.API_VERSION)
        offset = self.HEADER_LEN
        struct.pack_into('!ii', output, offset, -1, len(self._reqs))
        offset += 8
        for topic_name, partitions in iteritems(self._reqs):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name),
                             topic_name, len(partitions))
            offset += struct.calcsize(fmt)
            for pnum, offsets_before in iteritems(partitions):
                struct.pack_into('!iq', output, offset, pnum, offsets_before)
                offset += 12
        return output


OffsetPartitionResponse = namedtuple(
    'OffsetPartitionResponse',
    ['offset', 'err']
)


OffsetPartitionResponseV1 = namedtuple(
    'OffsetPartitionResponseV1',
    ['offset', 'timestamp', 'err']
)


class ListOffsetResponse(Response):
    """An offset response
    Specification::
        ListOffsetResponse => [TopicName [PartitionOffsets]]
          PartitionOffsets => Partition ErrorCode [Offset]
          Partition => int32
          ErrorCode => int16
          Offset => int64
    """
    API_VERSION = 0
    API_KEY = 2

    @classmethod
    def get_versions(cls):
        # XXX use ListOffsetResponseV1 after 0.10 message format is supported
        return {0: ListOffsetResponse, 1: ListOffsetResponse}

    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[S [ih [q] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.topics = {}
        for topic_name, partitions in response:
            self.topics[topic_name] = {}
            for partition in partitions:
                self.topics[topic_name][partition[0]] = OffsetPartitionResponse(
                    partition[2], partition[1])


class ListOffsetResponseV1(ListOffsetResponse):
    """
    Specification::
        ListOffsetResponse => [TopicName [PartitionOffsets]]
          PartitionOffsets => Partition ErrorCode Timestamp [Offset]
          Partition => int32
          ErrorCode => int16
          Timestamp => int64
          Offset => int64
    """
    API_VERSION = 1

    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[S [ihq [q] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.topics = {}
        for topic_name, partitions in response:
            self.topics[topic_name] = {}
            for partition in partitions:
                self.topics[topic_name][partition[0]] = OffsetPartitionResponseV1(
                    partition[3], partition[2], partition[1])



# - coding: utf-8 -
import struct
from collections import namedtuple, defaultdict

from .base import Request, Response
from ..utils import struct_helpers
from ..utils.compat import iteritems


class GroupCoordinatorRequest(Request):
    """A consumer metadata request
    Specification::
        GroupCoordinatorRequest => ConsumerGroup
            ConsumerGroup => string
    """
    API_KEY = 10

    def __init__(self, consumer_group):
        """Create a new group coordinator request"""
        self.consumer_group = consumer_group

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + len(self.consumer_group)
        return self.HEADER_LEN + 2 + len(self.consumer_group)

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        cglen = len(self.consumer_group)
        struct.pack_into('!h%ds' % cglen, output, self.HEADER_LEN, cglen,
                         self.consumer_group)
        return output


class GroupCoordinatorResponse(Response):
    """A group coordinator response
    Specification::
        GroupCoordinatorResponse => ErrorCode CoordinatorId CoordinatorHost CoordinatorPort
            ErrorCode => int16
            CoordinatorId => int32
            CoordinatorHost => string
            CoordinatorPort => int32
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'hiSi'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        error_code = response[0]
        if error_code != 0:
            self.raise_error(error_code, response)
        self.coordinator_id = response[1]
        self.coordinator_host = response[2]
        self.coordinator_port = response[3]


_PartitionOffsetCommitRequest = namedtuple(
    'PartitionOffsetCommitRequest',
    ['topic_name', 'partition_id', 'offset', 'timestamp', 'metadata']
)


class PartitionOffsetCommitRequest(_PartitionOffsetCommitRequest):
    """Offset commit request for a specific topic/partition
    :ivar topic_name: Name of the topic to look up
    :ivar partition_id: Id of the partition to look up
    :ivar offset:
    :ivar timestamp:
    :ivar metadata: arbitrary metadata that should be committed with this offset commit
    """
    pass


class OffsetCommitRequest(Request):
    """An offset commit request
    Specification::
        OffsetCommitRequest => ConsumerGroupId ConsumerGroupGenerationId ConsumerId [TopicName [Partition Offset TimeStamp Metadata]]
            ConsumerGroupId => string
            ConsumerGroupGenerationId => int32
            ConsumerId => string
            TopicName => string
            Partition => int32
            Offset => int64
            TimeStamp => int64
            Metadata => string
    """
    API_KEY = 8

    def __init__(self,
                 consumer_group,
                 consumer_group_generation_id,
                 consumer_id,
                 partition_requests=[]):
        """Create a new offset commit request
        :param partition_requests: Iterable of
            :class:`kafka.pykafka.protocol.PartitionOffsetCommitRequest` for
            this request
        """
        self.consumer_group = consumer_group
        self.consumer_group_generation_id = consumer_group_generation_id
        self.consumer_id = consumer_id
        self._reqs = defaultdict(dict)
        for t in partition_requests:
            self._reqs[t.topic_name][t.partition_id] = (t.offset,
                                                        t.timestamp,
                                                        t.metadata)

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + string size + consumer group size
        size = self.HEADER_LEN + 2 + len(self.consumer_group)
        # + generation id + string size + consumer_id size + array length
        size += 4 + 2 + len(self.consumer_id) + 4
        for topic, parts in iteritems(self._reqs):
            # topic name + len(parts)
            size += 2 + len(topic) + 4
            # partition + offset + timestamp => for each partition
            size += (4 + 8 + 8) * len(parts)
            # metadata => for each partition
            for partition, (_, _, metadata) in iteritems(parts):
                size += 2 + len(metadata)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output, api_version=1)
        offset = self.HEADER_LEN
        fmt = '!h%dsih%dsi' % (len(self.consumer_group), len(self.consumer_id))
        struct.pack_into(fmt, output, offset,
                         len(self.consumer_group), self.consumer_group,
                         self.consumer_group_generation_id,
                         len(self.consumer_id), self.consumer_id,
                         len(self._reqs))

        offset += struct.calcsize(fmt)
        for topic_name, partitions in iteritems(self._reqs):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name),
                             topic_name, len(partitions))
            offset += struct.calcsize(fmt)
            for pnum, (poffset, timestamp, metadata) in iteritems(partitions):
                fmt = '!iqq'
                struct.pack_into(fmt, output, offset,
                                 pnum, poffset, timestamp)
                offset += struct.calcsize(fmt)
                metalen = len(metadata) or -1
                fmt = '!h'
                pack_args = [fmt, output, offset, metalen]
                if metalen != -1:
                    fmt += '%ds' % metalen
                    pack_args = [fmt, output, offset, metalen, metadata]
                struct.pack_into(*pack_args)
                offset += struct.calcsize(fmt)
        return output


OffsetCommitPartitionResponse = namedtuple(
    'OffsetCommitPartitionResponse',
    ['err']
)


class OffsetCommitResponse(Response):
    """An offset commit response
    Specification::
        OffsetCommitResponse => [TopicName [Partition ErrorCode]]]
            TopicName => string
            Partition => int32
            ErrorCode => int16
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[S [ih ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.topics = {}
        for topic_name, partitions in response:
            self.topics[topic_name] = {}
            for partition in partitions:
                self.topics[topic_name][partition[0]] = OffsetCommitPartitionResponse(partition[1])


_PartitionOffsetFetchRequest = namedtuple(
    'PartitionOffsetFetchRequest',
    ['topic_name', 'partition_id']
)


class PartitionOffsetFetchRequest(_PartitionOffsetFetchRequest):
    """Offset fetch request for a specific topic/partition
    :ivar topic_name: Name of the topic to look up
    :ivar partition_id: Id of the partition to look up
    """
    pass


class OffsetFetchRequest(Request):
    """An offset fetch request
    Specification::
        OffsetFetchRequest => ConsumerGroup [TopicName [Partition]]
            ConsumerGroup => string
            TopicName => string
            Partition => int32
    """
    API_VERSION = 0
    API_KEY = 9

    @classmethod
    def get_versions(cls):
        return {0: OffsetFetchRequest, 1: OffsetFetchRequestV1, 2: OffsetFetchRequestV2}

    def __init__(self, consumer_group, partition_requests=[]):
        """Create a new offset fetch request
        :param partition_requests: Iterable of
            :class:`kafka.pykafka.protocol.PartitionOffsetFetchRequest` for
            this request
        """
        self.consumer_group = consumer_group
        self._reqs = defaultdict(list)
        for t in partition_requests:
            self._reqs[t.topic_name].append(t.partition_id)

    def _reqs_len(self):
        return len(self._reqs)

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + consumer group + len(topics)
        size = self.HEADER_LEN + 2 + len(self.consumer_group) + 4
        for topic, parts in iteritems(self._reqs):
            # topic name + len(parts)
            size += 2 + len(topic) + 4
            # partition => for each partition
            size += 4 * len(parts)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output, api_version=self.API_VERSION)
        offset = self.HEADER_LEN
        fmt = '!h%dsi' % len(self.consumer_group)
        struct.pack_into(fmt, output, offset,
                         len(self.consumer_group), self.consumer_group,
                         self._reqs_len())
        offset += struct.calcsize(fmt)
        for topic_name, partitions in iteritems(self._reqs):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name),
                             topic_name, len(partitions))
            offset += struct.calcsize(fmt)
            for pnum in partitions:
                fmt = '!i'
                struct.pack_into(fmt, output, offset, pnum)
                offset += struct.calcsize(fmt)
        return output


class OffsetFetchRequestV1(OffsetFetchRequest):
    API_VERSION = 1


class OffsetFetchRequestV2(OffsetFetchRequestV1):
    API_VERSION = 2

    def _reqs_len(self):
        # v2 allows a null array to select all topics
        return len(self._reqs) or -1


OffsetFetchPartitionResponse = namedtuple(
    'OffsetFetchPartitionResponse',
    ['offset', 'metadata', 'err']
)


class OffsetFetchResponse(Response):
    """An offset fetch response v0
    Specification::
    OffsetFetch Response (Version: 0) => [responses]
        responses => topic [partition_responses]
            topic => STRING
            partition_responses => partition offset metadata error_code
                partition => INT32
                offset => INT64
                metadata => NULLABLE_STRING
                error_code => INT16
    """
    API_VERSION = 0
    API_KEY = 9

    @classmethod
    def get_versions(cls):
        return {0: OffsetFetchResponse, 1: OffsetFetchResponseV1, 2: OffsetFetchResponseV2}

    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[S [iqSh ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        self._populate_partition_responses(response)

    def _populate_partition_responses(self, partition_responses):
        self.topics = {}
        for topic_name, partitions in partition_responses:
            self.topics[topic_name] = {}
            for partition in partitions:
                pres = OffsetFetchPartitionResponse(partition[1],
                                                    partition[2],
                                                    partition[3])
                self.topics[topic_name][partition[0]] = pres


class OffsetFetchResponseV1(OffsetFetchResponse):
    """An offset fetch response v1 (all the same as v0)
    Specification::
    OffsetFetch Response (Version: 1) => [responses]
        responses => topic [partition_responses]
            topic => STRING
            partition_responses => partition offset metadata error_code
                partition => INT32
                offset => INT64
                metadata => NULLABLE_STRING
                error_code => INT16
    """
    API_VERSION = 1


class OffsetFetchResponseV2(OffsetFetchResponseV1):
    """An offset fetch response v2
    Specification::
    OffsetFetch Response (Version: 2) => [responses] error_code
        responses => topic [partition_responses]
            topic => STRING
            partition_responses => partition offset metadata error_code
                partition => INT32
                offset => INT64
                metadata => NULLABLE_STRING
                error_code => INT16
        error_code => INT16 (new since v1)
    """
    API_VERSION = 2

    def __init__(self, buff):
        fmt = '[S [iqSh ] ] h'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        partition_responses, self.err = response
        self._populate_partition_responses(partition_responses)



from __future__ import division

import math
import mock
import os
import pkg_resources
import platform
import pytest
import time
import threading
import unittest2
from uuid import uuid4

from kazoo.client import KazooClient
try:
    import gevent
except ImportError:
    gevent = None

from pykafka import KafkaClient
from pykafka.balancedconsumer import BalancedConsumer, OffsetType
from pykafka.exceptions import ConsumerStoppedException
from pykafka.managedbalancedconsumer import ManagedBalancedConsumer
from pykafka.membershipprotocol import (GroupMembershipProtocol, RoundRobinProtocol,
                                        RangeProtocol)
from pykafka.test.utils import get_cluster, stop_cluster
from pykafka.utils.compat import range, iterkeys, iteritems


kafka_version_string = os.environ.get('KAFKA_VERSION', '0.8')
kafka_version = pkg_resources.parse_version(kafka_version_string)
version_09 = pkg_resources.parse_version("0.9.0.0")


class TestBalancedConsumer(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._consumer_timeout = 2000
        cls._mock_consumer, _ = TestBalancedConsumer.buildMockConsumer(timeout=cls._consumer_timeout)

    @classmethod
    def buildMockConsumer(self, consumer_group=b'testgroup', num_partitions=10,
                          num_participants=1, timeout=2000):
        topic = mock.Mock()
        topic.name = 'testtopic'
        topic.partitions = {}
        for k in range(num_partitions):
            part = mock.Mock(name='part-{part}'.format(part=k))
            part.id = k
            part.topic = topic
            part.leader = mock.Mock()
            part.leader.id = k % num_participants
            topic.partitions[k] = part

        cluster = mock.MagicMock()
        zk = mock.MagicMock()
        return BalancedConsumer(topic, cluster, consumer_group,
                                zookeeper=zk, auto_start=False, use_rdkafka=False,
                                consumer_timeout_ms=timeout), topic

    def test_unicode_consumer_group(self):
        consumer, _ = self.buildMockConsumer(consumer_group=u'testgroup')

    def test_consume_returns(self):
        """Ensure that consume() returns in the amount of time it's supposed to
        """
        self._mock_consumer._setup_internal_consumer(start=False)
        self._mock_consumer._consumer._partitions_by_id = {1: "dummy"}
        self._mock_consumer._running = True
        start = time.time()
        self._mock_consumer.consume()
        self.assertEqual(int(time.time() - start), int(self._consumer_timeout / 1000))

    def test_consume_graceful_stop(self):
        """Ensure that stopping a consumer while consuming from Kafka does not
        end in an infinite loop when timeout is not used.
        """
        consumer, _ = self.buildMockConsumer(timeout=-1)
        consumer._setup_internal_consumer(start=False)
        consumer._consumer._partitions_by_id = {1: "dummy"}

        consumer.stop()
        with self.assertRaises(ConsumerStoppedException):
            consumer.consume()

    def _test_decide_partitions(self, membership_protocol):
        for i in range(100):
            num_participants = i + 1
            num_partitions = 100 - i
            participants = sorted(['test-debian:{p}'.format(p=p)
                                   for p in range(num_participants)])
            cns, topic = self.buildMockConsumer(num_partitions=num_partitions,
                                                num_participants=num_participants)
            cns._membership_protocol = membership_protocol
            assigned_parts = []
            for consumer_id in participants:
                partitions = cns._membership_protocol.decide_partitions(
                    participants, topic.partitions, consumer_id)
                assigned_parts.extend(partitions)

                remainder_ppc = num_partitions % num_participants
                idx = participants.index(consumer_id)
                parts_per_consumer = math.floor(num_partitions / num_participants)
                num_parts = parts_per_consumer + (0 if (idx + 1 > remainder_ppc) else 1)

                self.assertEqual(len(partitions), int(num_parts))

            # Validate all partitions were assigned once and only once
            all_partitions = sorted(topic.partitions.values(), key=lambda x: x.id)
            assigned_parts = sorted(assigned_parts, key=lambda x: x.id)
            self.assertListEqual(assigned_parts, all_partitions)

    def test_decide_partitions_range(self):
        self._test_decide_partitions(RangeProtocol)

    def test_decide_partitions_roundrobin(self):
        self._test_decide_partitions(RoundRobinProtocol)


class TestManagedBalancedConsumer(TestBalancedConsumer):
    @classmethod
    def buildMockConsumer(self, consumer_group=b'testgroup', num_partitions=10,
                          num_participants=1, timeout=2000):
        topic = mock.Mock()
        topic.name = 'testtopic'
        topic.partitions = {}
        for k in range(num_partitions):
            part = mock.Mock(name='part-{part}'.format(part=k))
            part.id = k
            part.topic = topic
            part.leader = mock.Mock()
            part.leader.id = k % num_participants
            topic.partitions[k] = part

        cluster = mock.MagicMock()
        cns = ManagedBalancedConsumer(topic, cluster, consumer_group,
                                      auto_start=False, use_rdkafka=False,
                                      consumer_timeout_ms=timeout)
        cns._group_coordinator = mock.MagicMock()
        return cns, topic


class BalancedConsumerIntegrationTests(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False
    MANAGED_CONSUMER = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = uuid4().hex.encode()
        cls.n_partitions = 3
        cls.kafka.create_topic(cls.topic_name, cls.n_partitions, 2)
        cls.total_msgs = 1000
        cls.client = KafkaClient(cls.kafka.brokers,
                                 use_greenlets=cls.USE_GEVENT,
                                 broker_version=kafka_version_string)
        cls.prod = cls.client.topics[cls.topic_name].get_producer(
            min_queued_messages=1
        )
        for i in range(cls.total_msgs):
            cls.prod.produce('msg {num}'.format(num=i).encode())

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def get_zk(self):
        if not self.USE_GEVENT:
            return KazooClient(self.kafka.zookeeper)

        from kazoo.handlers.gevent import SequentialGeventHandler

        return KazooClient(self.kafka.zookeeper, handler=SequentialGeventHandler())

    def get_balanced_consumer(self, consumer_group, **kwargs):
        if self.MANAGED_CONSUMER:
            kwargs.pop("zookeeper", None)
            kwargs.pop("zookeeper_connect", None)
        return self.client.topics[self.topic_name].get_balanced_consumer(
            consumer_group,
            managed=self.MANAGED_CONSUMER,
            **kwargs
        )

    def test_extra_consumer(self):
        """Ensure proper operation of "extra" consumers in a group

        An "extra" consumer is the N+1th member of a consumer group consuming a topic
        of N partitions, and any consumer beyond the N+1th.
        """
        group = b"test_extra_consumer"
        extras = 1

        def verify_extras(consumers, extras_count):
            messages = [c.consume() for c in consumers]
            successes = [a for a in messages if a is not None]
            nones = [a for a in messages if a is None]
            attempts = 0
            while len(nones) != extras_count and attempts < 5:
                messages = [c.consume() for c in consumers]
                successes = [a for a in messages if a is not None]
                nones = [a for a in messages if a is None]
                attempts += 1
            self.assertEqual(len(nones), extras_count)
            self.assertEqual(len(successes), self.n_partitions)

        try:
            consumers = [self.get_balanced_consumer(group, consumer_timeout_ms=5000)
                         for i in range(self.n_partitions + extras)]
            verify_extras(consumers, extras)

            # when one consumer stops, the extra should pick up its partitions
            removed = consumers[:extras]
            for consumer in removed:
                consumer.stop()
            consumers = [a for a in consumers if a not in removed]
            self.wait_for_rebalancing(*consumers)
            self.assertEqual(len(consumers), self.n_partitions)
            verify_extras(consumers, 0)

            # added "extra" consumers should idle
            for i in range(extras):
                consumers.append(self.get_balanced_consumer(group,
                                                            consumer_timeout_ms=5000))
            self.wait_for_rebalancing(*consumers)
            verify_extras(consumers, extras)
        finally:
            for consumer in consumers:
                try:
                    consumer.stop()
                except:
                    pass

    # weird name to ensure test execution order, because there is an unintended
    # interdependency between test_consume_latest and other tests
    def test_a_rebalance_unblock_event(self):
        """Adding a new consumer instance to a group should release
        blocking consume() call of any existing consumer instance(s).

        https://github.com/Parsely/pykafka/issues/701
        """
        if self.USE_GEVENT:
            pytest.skip("Unresolved failure")
        group = b'test_rebalance'
        consumer_a = self.get_balanced_consumer(group, consumer_timeout_ms=-1)

        # consume all msgs to block the consume() call
        count = 0
        for _ in consumer_a:
            count += 1
            if count == self.total_msgs:
                break

        consumer_a_thread = threading.Thread(target=consumer_a.consume)
        consumer_a_thread.start()

        consumer_b = self.get_balanced_consumer(group, consumer_timeout_ms=-1)
        consumer_b_thread = threading.Thread(target=consumer_b.consume)
        consumer_b_thread.start()

        consumer_a_thread.join(30)
        consumer_b_thread.join(30)

        # consumer thread would die in case of any rebalancing errors
        self.assertTrue(consumer_a_thread.is_alive() and consumer_b_thread.is_alive())

    def test_rebalance_callbacks(self):
        def on_rebalance(cns, old_partition_offsets, new_partition_offsets):
            self.assertTrue(len(new_partition_offsets) > 0)
            self.assigned_called = True
            for id_ in iterkeys(new_partition_offsets):
                new_partition_offsets[id_] = self.offset_reset
            return new_partition_offsets

        self.assigned_called = False
        self.offset_reset = 50
        try:
            consumer_group = b'test_rebalance_callbacks'
            consumer_a = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                post_rebalance_callback=on_rebalance,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)
            self.wait_for_rebalancing(consumer_a, consumer_b)
            self.assertTrue(self.assigned_called)
            for _, offset in iteritems(consumer_a.held_offsets):
                self.assertEqual(offset, self.offset_reset)
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_rebalance_callbacks_surfaces_errors(self):
        def on_rebalance(cns, old_partition_offsets, new_partition_offsets):
            raise ValueError("BAD CALLBACK")

        self.assigned_called = False
        self.offset_reset = 50
        try:
            consumer_group = b'test_rebalance_callbacks_error'
            consumer_a = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                post_rebalance_callback=on_rebalance,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                consumer_group,
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)

            with pytest.raises(ValueError) as ex:
                self.wait_for_rebalancing(consumer_a, consumer_b)
                assert 'BAD CALLBACK' in str(ex.value)

        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_consume_earliest(self):
        try:
            consumer_a = self.get_balanced_consumer(
                b'test_consume_earliest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                b'test_consume_earliest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.EARLIEST,
                use_rdkafka=self.USE_RDKAFKA)

            # Consume from both a few times
            messages = [consumer_a.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)
            messages = [consumer_b.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)

            # Validate they aren't sharing partitions
            self.assertSetEqual(
                consumer_a._partitions & consumer_b._partitions,
                set()
            )

            # Validate all partitions are here
            self.assertSetEqual(
                consumer_a._partitions | consumer_b._partitions,
                set(self.client.topics[self.topic_name].partitions.values())
            )
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_consume_latest(self):
        try:
            consumer_a = self.get_balanced_consumer(
                b'test_consume_latest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.LATEST,
                use_rdkafka=self.USE_RDKAFKA)
            consumer_b = self.get_balanced_consumer(
                b'test_consume_latest',
                zookeeper_connect=self.kafka.zookeeper,
                auto_offset_reset=OffsetType.LATEST,
                use_rdkafka=self.USE_RDKAFKA)

            # Make sure we're done before producing more messages:
            self.wait_for_rebalancing(consumer_a, consumer_b)

            # Since we are consuming from the latest offset,
            # produce more messages to consume.
            for i in range(10):
                self.prod.produce('msg {num}'.format(num=i).encode())

            # Consume from both a few times
            messages = [consumer_a.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)
            messages = [consumer_b.consume() for i in range(1)]
            self.assertTrue(len(messages) == 1)

            # Validate they aren't sharing partitions
            self.assertSetEqual(
                consumer_a._partitions & consumer_b._partitions,
                set()
            )

            # Validate all partitions are here
            self.assertSetEqual(
                consumer_a._partitions | consumer_b._partitions,
                set(self.client.topics[self.topic_name].partitions.values())
            )
        finally:
            try:
                consumer_a.stop()
                consumer_b.stop()
            except:
                pass

    def test_external_kazoo_client(self):
        """Run with pre-existing KazooClient instance

        This currently doesn't assert anything, it just rules out any trivial
        exceptions in the code path that uses an external KazooClient
        """
        if self.MANAGED_CONSUMER:
            pytest.skip("Managed consumer doesn't use zookeeper")
        zk = KazooClient(self.kafka.zookeeper)
        zk.start()

        consumer = self.get_balanced_consumer(
            b'test_external_kazoo_client',
            zookeeper=zk,
            consumer_timeout_ms=10,
            use_rdkafka=self.USE_RDKAFKA)
        [msg for msg in consumer]
        consumer.stop()

    def test_no_partitions(self):
        """Ensure a consumer assigned no partitions doesn't fail"""

        def _decide_dummy(participants, partitions, consumer_id):
            return set()
        consumer = self.get_balanced_consumer(
            b'test_no_partitions',
            zookeeper_connect=self.kafka.zookeeper,
            auto_start=False,
            consumer_timeout_ms=50,
            use_rdkafka=self.USE_RDKAFKA)

        consumer._membership_protocol = GroupMembershipProtocol(
            consumer._membership_protocol.protocol_type,
            consumer._membership_protocol.protocol_name,
            consumer._membership_protocol.metadata,
            _decide_dummy
        )
        consumer.start()
        res = consumer.consume()
        self.assertEqual(res, None)
        self.assertTrue(consumer._running)
        # check that stop() succeeds (cf #313 and #392)
        consumer.stop()

    def test_zk_conn_lost(self):
        """Check we restore zookeeper nodes correctly after connection loss

        See also github issue #204.
        """
        if self.MANAGED_CONSUMER:
            pytest.skip("Managed consumer doesn't use zookeeper")
        check_partitions = lambda c: c._get_held_partitions() == c._partitions
        zk = self.get_zk()
        zk.start()
        try:
            consumer_group = b'test_zk_conn_lost'

            consumer = self.get_balanced_consumer(consumer_group,
                                                  zookeeper=zk,
                                                  use_rdkafka=self.USE_RDKAFKA)
            self.assertTrue(check_partitions(consumer))
            with consumer._rebalancing_lock:
                zk.stop()  # expires session, dropping all our nodes

            # Start a second consumer on a different zk connection
            other_consumer = self.get_balanced_consumer(
                consumer_group, use_rdkafka=self.USE_RDKAFKA)

            # Slightly contrived: we'll grab a lock to keep _rebalance() from
            # starting when we restart the zk connection (restart triggers a
            # rebalance), so we can confirm the expected discrepancy between
            # the (empty) set of partitions on zk and the set in the internal
            # consumer:
            with consumer._rebalancing_lock:
                zk.start()
                self.assertFalse(check_partitions(consumer))

            # Finally, confirm that _rebalance() resolves the discrepancy:
            self.wait_for_rebalancing(consumer, other_consumer)
            self.assertTrue(check_partitions(consumer))
            self.assertTrue(check_partitions(other_consumer))
        finally:
            try:
                consumer.stop()
                other_consumer.stop()
                zk.stop()
            except:
                pass

    def wait_for_rebalancing(self, *balanced_consumers):
        """Test helper that loops while rebalancing is ongoing

        Needs to be given all consumer instances active in a consumer group.
        Waits for up to 100 seconds, which should be enough for even a very
        oversubscribed test cluster.
        """
        for _ in range(500):
            n_parts = [len(cons.partitions) for cons in balanced_consumers]
            if (max(n_parts) - min(n_parts) <= 1
                    and sum(n_parts) == self.n_partitions):
                break
            else:
                balanced_consumers[0]._cluster.handler.sleep(.2)
            # check for failed consumers (there'd be no point waiting anymore)
            [cons._raise_worker_exceptions() for cons in balanced_consumers]
        else:
            raise AssertionError("Rebalancing failed")


@pytest.mark.skipif(platform.python_implementation() == "PyPy" or gevent is None,
                    reason="Unresolved crashes")
class BalancedConsumerGEventIntegrationTests(BalancedConsumerIntegrationTests):
    USE_GEVENT = True


@pytest.mark.skipif(kafka_version < version_09,
                    reason="Managed consumer unsupported until 0.9")
class ManagedBalancedConsumerIntegrationTests(BalancedConsumerIntegrationTests):
    MANAGED_CONSUMER = True


@pytest.mark.skipif(platform.python_implementation() == "PyPy" or
                    kafka_version < version_09 or gevent is None,
                    reason="Unresolved crashes")
class ManagedBalancedConsumerGEventIntegrationTests(BalancedConsumerIntegrationTests):
    MANAGED_CONSUMER = True
    USE_GEVENT = True


if __name__ == "__main__":
    unittest2.main()



from contextlib import contextmanager
import datetime as dt
import json
import mock
import os
import platform
import pytest
import time
import threading
import unittest2
from uuid import uuid4

try:
    import gevent
except ImportError:
    gevent = None

try:
    from pykafka.rdkafka import _rd_kafka  # noqa
    RDKAFKA = True
except ImportError:
    RDKAFKA = False  # C extension not built

from pykafka import KafkaClient
from pykafka.simpleconsumer import OwnedPartition, OffsetType
from pykafka.test.utils import get_cluster, stop_cluster
from pykafka.utils.compat import range, iteritems, get_string


kafka_version = os.environ.get('KAFKA_VERSION', '0.8.0')


class TestSimpleConsumer(unittest2.TestCase):
    maxDiff = None
    USE_RDKAFKA = False
    USE_GEVENT = False

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = uuid4().hex.encode()
        cls.kafka.create_topic(cls.topic_name, 3, 2)

        cls.total_msgs = 1000
        cls.client = KafkaClient(cls.kafka.brokers, broker_version=kafka_version)
        cls.prod = cls.client.topics[cls.topic_name].get_producer(
            min_queued_messages=1
        )
        cls.after_earliest = dt.datetime.now() + dt.timedelta(seconds=1)
        for i in range(cls.total_msgs):
            cls.prod.produce('msg {i}'.format(i=i).encode())

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    @contextmanager
    def _get_simple_consumer(self, **kwargs):
        topic = self.client.topics[self.topic_name]
        consumer = topic.get_simple_consumer(
            use_rdkafka=self.USE_RDKAFKA, **kwargs)
        try:
            yield consumer
        finally:
            consumer.stop()

    def test_consume(self):
        """Test consuming all messages in topic"""
        # This uses a fairly long timeout to allow the test to pass on an
        # oversubscribed test cluster
        with self._get_simple_consumer(consumer_timeout_ms=30000) as consumer:
            count = 0
            for msg in consumer:
                self.assertIsNotNone(msg.value)
                count += 1
                if count == self.total_msgs:
                    # We don't want to wait for StopIteration, given the long
                    # timeout set above
                    break
            self.assertEquals(count, self.total_msgs)

    def test_unblock_event(self):
        """A call to consume() should return when unblock_event is set()

        To block consume() method indefinitely, all messages in the topic need to
        be consumed and the timeout should be set to -1.
        """
        topic = self.client.topics[self.topic_name]
        consumer = topic.get_simple_consumer(use_rdkafka=self.USE_RDKAFKA, consumer_timeout_ms=-1)

        # Consume all messages in the topic
        count = 0
        for _ in consumer:
            count += 1
            if count == self.total_msgs:
                break

        unblock_event = consumer._cluster.handler.Event()
        consume_thread = threading.Thread(target=consumer.consume, kwargs={'unblock_event': unblock_event})
        consume_thread.start()
        unblock_event.set()
        consume_thread.join(30)
        self.assertFalse(consume_thread.is_alive())

    @staticmethod
    def _convert_offsets(offset_responses):
        """Helper function to translate Offset(Fetch)PartitionResponse

        Calls like consumer.fetch_offsets() and earliest_available_offsets()
        return lists of OffsetPartitionResponses.  These hold the next offset
        to be consumed, whereas consumer.held_offsets returns the latest
        consumed offset.  This translates them to facilitate comparisons.
        """
        if isinstance(offset_responses, dict):
            offset_responses = iteritems(offset_responses)
        f1 = lambda off: OffsetType.EARLIEST if off == 0 else off - 1  # noqa
        f2 = lambda off: off[0] if isinstance(off, list) else off  # noqa
        return {partition_id: f1(f2(offset_response.offset))
                for partition_id, offset_response in offset_responses}

    def test_offset_commit(self):
        """Check fetched offsets match pre-commit internal state"""
        with self._get_simple_consumer(
                consumer_group=b'test_offset_commit') as consumer:
            [consumer.consume() for _ in range(100)]
            offsets_committed = consumer.held_offsets
            consumer.commit_offsets()

            offsets_fetched = self._convert_offsets(consumer.fetch_offsets())
            self.assertEquals(offsets_fetched, offsets_committed)

    def test_offset_commit_override(self):
        """Check fetched offsets match committed offsets"""
        with self._get_simple_consumer(
                consumer_group=b'test_offset_commit') as consumer:
            [consumer.consume() for _ in range(100)]
            offset = 69
            offsets_committed = [(p, offset) for p in consumer.partitions.values()]
            consumer.commit_offsets(partition_offsets=offsets_committed)

            offsets_fetched = self._convert_offsets(consumer.fetch_offsets())
            offsets_committed = {p.id: offset - 1 for p in consumer.partitions.values()}
            self.assertEquals(offsets_fetched, offsets_committed)

    def test_offset_resume(self):
        """Check resumed internal state matches committed offsets"""
        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume') as consumer:
            [consumer.consume() for _ in range(100)]
            offsets_committed = consumer.held_offsets
            consumer.commit_offsets()

        with self._get_simple_consumer(
                consumer_group=b'test_offset_resume') as consumer:
            self.assertEquals(consumer.held_offsets, offsets_committed)

    def test_reset_offset_on_start(self):
        """Try starting from LATEST and EARLIEST offsets"""
        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.EARLIEST,
                reset_offset_on_start=True) as consumer:
            earliest_offs = self._convert_offsets(
                consumer.topic.earliest_available_offsets())
            self.assertEquals(earliest_offs, consumer.held_offsets)
            self.assertIsNotNone(consumer.consume())

        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.LATEST,
                reset_offset_on_start=True,
                consumer_timeout_ms=500) as consumer:
            latest_offs = self._convert_offsets(
                consumer.topic.latest_available_offsets())
            self.assertEquals(latest_offs, consumer.held_offsets)
            self.assertIsNone(consumer.consume(block=False))

        difference = sum(latest_offs[i] - earliest_offs[i]
                         if earliest_offs[i] >= 0 else latest_offs[i] + 1
                         if latest_offs[i] >= 0 else 0
                         for i in latest_offs)
        self.assertEqual(difference, self.total_msgs)

    def test_reset_offsets_timestamp(self):
        """Test resetting to user-provided timestamps"""
        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.EARLIEST) as consumer:
            # Find us a non-empty partition "target_part"
            part_id, latest_offset = next(
                (p, res.offset[0])
                for p, res in consumer.topic.latest_available_offsets().items()
                if res.offset[0] > 0)
            target_part = consumer.partitions[part_id]

            # Set all other partitions to LATEST, to ensure that any consume()
            # calls read from target_part
            partition_offsets = {
                p: OffsetType.LATEST for p in consumer.partitions.values()}

            partition_offsets[target_part] = self.after_earliest
            consumer.reset_offsets(partition_offsets.items())

            # expect EARLIEST here since our test partition has a single log segment
            self.assertEqual(consumer.held_offsets[part_id], OffsetType.EARLIEST)
            msg = consumer.consume()
            self.assertEqual(msg.offset, 0)

    def test_reset_offsets(self):
        """Test resetting to user-provided offsets"""
        with self._get_simple_consumer(
                auto_offset_reset=OffsetType.EARLIEST) as consumer:
            # Find us a non-empty partition "target_part"
            part_id, latest_offset = next(
                (p, res.offset[0])
                for p, res in consumer.topic.latest_available_offsets().items()
                if res.offset[0] > 0)
            target_part = consumer.partitions[part_id]

            # Set all other partitions to LATEST, to ensure that any consume()
            # calls read from target_part
            partition_offsets = {
                p: OffsetType.LATEST for p in consumer.partitions.values()}

            new_offset = latest_offset - 5
            partition_offsets[target_part] = new_offset
            consumer.reset_offsets(partition_offsets.items())

            self.assertEqual(consumer.held_offsets[part_id], new_offset)
            msg = consumer.consume()
            self.assertEqual(msg.offset, new_offset + 1)

            # Invalid offsets should get overwritten as per auto_offset_reset
            partition_offsets[target_part] = latest_offset + 5  # invalid!
            consumer.reset_offsets(partition_offsets.items())

            # SimpleConsumer's fetcher thread will detect the invalid offset
            # and reset it immediately.  RdKafkaSimpleConsumer however will
            # only get to write the valid offset upon a call to consume():
            msg = consumer.consume()
            expected_offset = target_part.earliest_available_offset()
            self.assertEqual(msg.offset, expected_offset)
            self.assertEqual(consumer.held_offsets[part_id], expected_offset)

    @pytest.mark.xfail
    def test_update_cluster(self):
        """Check that the consumer can initiate cluster updates"""
        if self.USE_RDKAFKA:
            pytest.skip("Unresolved crashes")
        with self._get_simple_consumer() as consumer:
            self.assertIsNotNone(consumer.consume())

            for broker in self.client.brokers.values():
                broker._connection.disconnect()

            # The consumer fetcher thread should prompt broker reconnection
            t_start = time.time()
            timeout = 40. if self.USE_GEVENT else 20.
            try:
                for broker in self.client.brokers.values():
                    while not broker._connection.connected:
                        time.sleep(.1)
                        self.assertTrue(time.time() - t_start < timeout,
                                        msg="Broker reconnect failed.")
            finally:
                # Make sure further tests don't get confused
                consumer._update()
            # If the fetcher thread fell over during the cluster update
            # process, we'd get an exception here:
            self.assertIsNotNone(consumer.consume())

    @pytest.mark.xfail
    def test_consumer_lag(self):
        """Ensure that after consuming the entire topic, lag is 0"""
        with self._get_simple_consumer(consumer_group=b"test_lag_group",
                                       consumer_timeout_ms=1000) as consumer:
            while True:
                message = consumer.consume()
                if message is None:
                    break
            consumer.commit_offsets()
            latest_offsets = {p_id: res.offset[0]
                              for p_id, res
                              in iteritems(consumer.topic.latest_available_offsets())}
            current_offsets = {p_id: res.offset for p_id, res in consumer.fetch_offsets()}
            self.assertEqual(current_offsets, latest_offsets)


@pytest.mark.skipif(platform.python_implementation() == "PyPy" or gevent is None,
                    reason="Unresolved crashes")
class TestGEventSimpleConsumer(TestSimpleConsumer):
    USE_GEVENT = True


class TestOwnedPartition(unittest2.TestCase):
    def test_partition_saves_offset(self):
        offset = 20
        msgval = "test"
        partition = mock.MagicMock()
        op = OwnedPartition(partition)
        op.next_offset = offset

        message = mock.Mock()
        message.value = msgval
        message.offset = offset

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 1)
        ret_message = op.consume()
        self.assertEqual(op.last_offset_consumed, message.offset)
        self.assertEqual(op.next_offset, message.offset + 1)
        self.assertNotEqual(ret_message, None)
        self.assertEqual(ret_message.value, msgval)

    def test_partition_rejects_old_message(self):
        last_offset = 400
        op = OwnedPartition(None)
        op.last_offset_consumed = last_offset

        message = mock.Mock()
        message.value = "test"
        message.offset = 20

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 0)
        op.consume()
        self.assertEqual(op.last_offset_consumed, last_offset)

    def test_compacted_topic_partition_rejects_old_message_after_initial(self):
        last_offset = 400
        message1 = mock.Mock()
        message1.value = "first-test"
        message1.partition_id = 0
        message1.offset = last_offset

        partition = mock.MagicMock()
        partition.id = 0
        op = OwnedPartition(partition, compacted_topic=True)
        op.enqueue_messages([message1])
        self.assertEqual(op.message_count, 1)
        op.consume()
        self.assertEqual(op.message_count, 0)
        self.assertEqual(op.last_offset_consumed, last_offset)

        message2 = mock.Mock()
        message2.value = "test"
        message2.partition_id = 0
        message2.offset = 20

        op.enqueue_messages([message2])
        self.assertEqual(op.message_count, 0)
        op.consume()
        self.assertEqual(op.last_offset_consumed, last_offset)

    def test_partition_consume_empty_queue(self):
        op = OwnedPartition(None)

        message = op.consume()
        self.assertEqual(message, None)

    def test_partition_offset_commit_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)
        op.last_offset_consumed = 200

        request = op.build_offset_commit_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)
        self.assertEqual(request.offset, op.last_offset_consumed + 1)
        parsed_metadata = json.loads(get_string(request.metadata))
        self.assertEqual(parsed_metadata["consumer_id"], '')
        self.assertTrue(bool(parsed_metadata["hostname"]))

    def test_partition_offset_fetch_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)

        request = op.build_offset_fetch_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)

    def test_partition_offset_counters(self):
        res = mock.Mock()
        res.offset = 400

        op = OwnedPartition(None)
        op.set_offset(res.offset)

        self.assertEqual(op.last_offset_consumed, res.offset)
        self.assertEqual(op.next_offset, res.offset + 1)


if __name__ == "__main__":
    unittest2.main()



# - coding: utf-8 -
import struct
from collections import namedtuple

from .base import Request, Response
from ..utils import struct_helpers


class MetadataRequest(Request):
    """Metadata Request
    Specification::
        MetadataRequest => [TopicName]
            TopicName => string
    """
    API_VERSION = 0
    API_KEY = 3

    @classmethod
    def get_versions(cls):
        return {0: MetadataRequest, 1: MetadataRequestV1, 2: MetadataRequestV2,
                3: MetadataRequestV3, 4: MetadataRequestV4, 5: MetadataRequestV5}

    def __init__(self, topics=None, *kwargs):
        """Create a new MetadataRequest
        :param topics: Topics to query. Leave empty for all available topics.
        """
        self.topics = topics or []

    def __len__(self):
        """Length of the serialized message, in bytes"""
        return self.HEADER_LEN + 4 + sum(len(t) + 2 for t in self.topics)

    def _topics_len(self):
        return len(self.topics)

    def _serialize(self):
        output = bytearray(len(self))
        self._write_header(output, api_version=self.API_VERSION)
        struct.pack_into('!i', output, self.HEADER_LEN, self._topics_len())
        offset = self.HEADER_LEN + 4
        for t in self.topics:
            tlen = len(t)
            struct.pack_into('!h%ds' % tlen, output, offset, tlen, t)
            offset += 2 + tlen
        return output, offset

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output, _ = self._serialize()
        return output


class MetadataRequestV1(MetadataRequest):
    API_VERSION = 1

    def _topics_len(self):
        # v1 and higher require a null array, not an empty array, to select all topics
        return len(self.topics) or -1


class MetadataRequestV2(MetadataRequestV1):
    API_VERSION = 2


class MetadataRequestV3(MetadataRequestV2):
    API_VERSION = 3


class MetadataRequestV4(MetadataRequestV3):
    """Metadata Request
    Specification::
    Metadata Request (Version: 4) => [topics] allow_auto_topic_creation
        topics => STRING
        allow_auto_topic_creation => BOOLEAN
    """
    API_VERSION = 4

    def __init__(self, topics=None, allow_topic_autocreation=True):
        """Create a new MetadataRequest
        :param topics: Topics to query. Leave empty for all available topics.
        :param allow_topic_autocreation: If this and the broker config
            'auto.create.topics.enable' are true, topics that don't exist will be created
            by the broker. Otherwise, no topics will be created by the broker.
        """
        super(MetadataRequestV4, self).__init__(topics=topics)
        self.allow_topic_autocreation = allow_topic_autocreation

    def __len__(self):
        return super(MetadataRequestV4, self).__len__() + 1

    def get_bytes(self):
        output, offset = self._serialize()
        struct.pack_into('!b', output, offset, self.allow_topic_autocreation)
        return output


class MetadataRequestV5(MetadataRequestV4):
    API_VERSION = 5


BrokerMetadata = namedtuple('BrokerMetadata', ['id', 'host', 'port'])
BrokerMetadataV1 = namedtuple('BrokerMetadataV1', ['id', 'host', 'port', 'rack'])
TopicMetadata = namedtuple('TopicMetadata', ['name', 'partitions', 'err'])
TopicMetadataV1 = namedtuple('TopicMetadataV1', ['name', 'is_internal', 'partitions',
                                                 'err'])
PartitionMetadata = namedtuple('PartitionMetadata',
                               ['id', 'leader', 'replicas', 'isr', 'err'])
PartitionMetadataV5 = namedtuple('PartitionMetadataV5',
                                 ['id', 'leader', 'replicas', 'isr', 'offline_replicas',
                                  'err'])


class MetadataResponse(Response):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 0) => [brokers] [topic_metadata]
        brokers => node_id host port
            node_id => INT32
            host => STRING
            port => INT32
        topic_metadata => error_code topic [partition_metadata]
            error_code => INT16
            topic => STRING
            partition_metadata => error_code partition leader [replicas] [isr]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
    """
    API_KEY = 3

    @classmethod
    def get_versions(cls):
        return {0: MetadataResponse, 1: MetadataResponseV1, 2: MetadataResponseV2,
                3: MetadataResponseV3, 4: MetadataResponseV4, 5: MetadataResponseV5}

    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = '[iSi] [hS [hii [i] [i] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        broker_info, topics = response
        self._populate(broker_info, topics)

    def _populate(self,
                  broker_info,
                  topics,
                  controller_id=None,
                  cluster_id=None,
                  throttle_time_ms=0):
        self.throttle_time_ms = throttle_time_ms
        self._build_broker_metas(broker_info)
        self.cluster_id = cluster_id
        self.controller_id = controller_id
        self._build_topic_metas(topics)

    def _build_topic_metas(self, topics):
        self.topics = {}
        for (err, name, partitions) in topics:
            self.topics[name] = TopicMetadata(name,
                                              self._build_partition_metas(partitions),
                                              err)

    def _build_partition_metas(self, partitions):
        part_metas = {}
        for (p_err, id_, leader, replicas, isr) in partitions:
            part_metas[id_] = PartitionMetadata(id_, leader, replicas,
                                                isr, p_err)
        return part_metas

    def _build_broker_metas(self, broker_info):
        self.brokers = {}
        for (id_, host, port) in broker_info:
            self.brokers[id_] = BrokerMetadata(id_, host, port)


class MetadataResponseV1(MetadataResponse):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 1) => [brokers] controller_id [topic_metadata]
        brokers => node_id host port rack
            node_id => INT32
            host => STRING
            port => INT32
            rack => NULLABLE_STRING  (new since v0)
        controller_id => INT32  (new since v0)
        topic_metadata => error_code topic is_internal [partition_metadata]
            error_code => INT16
            topic => STRING
            is_internal => BOOLEAN  (new since v0)
            partition_metadata => error_code partition leader [replicas] [isr]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
    """
    def __init__(self, buff):
        fmt = '[iSiS] i [hSb [hii [i] [i] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        broker_info, controller_id, topics = response
        self._populate(broker_info, topics, controller_id=controller_id)

    def _build_topic_metas(self, topics):
        self.topics = {}
        for (err, name, is_internal, partitions) in topics:
            self.topics[name] = TopicMetadataV1(name,
                                                is_internal,
                                                self._build_partition_metas(partitions),
                                                err)

    def _build_broker_metas(self, broker_info):
        self.brokers = {}
        for (id_, host, port, rack) in broker_info:
            self.brokers[id_] = BrokerMetadataV1(id_, host, port, rack)


class MetadataResponseV2(MetadataResponseV1):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 2) => [brokers] cluster_id controller_id [topic_metadata]
        brokers => node_id host port rack
            node_id => INT32
            host => STRING
            port => INT32
            rack => NULLABLE_STRING
        cluster_id => NULLABLE_STRING  (new since v1)
        controller_id => INT32
        topic_metadata => error_code topic is_internal [partition_metadata]
            error_code => INT16
            topic => STRING
            is_internal => BOOLEAN
            partition_metadata => error_code partition leader [replicas] [isr]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
    """
    def __init__(self, buff):
        fmt = '[iSiS] Si [hSb [hii [i] [i] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        broker_info, cluster_id, controller_id, topics = response
        self._populate(broker_info, topics, controller_id=controller_id,
                       cluster_id=cluster_id)


class MetadataResponseV3(MetadataResponseV2):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 3) => throttle_time_ms [brokers] cluster_id controller_id [topic_metadata]
        throttle_time_ms => INT32  (new since v2)
        brokers => node_id host port rack
            node_id => INT32
            host => STRING
            port => INT32
            rack => NULLABLE_STRING
        cluster_id => NULLABLE_STRING
        controller_id => INT32
        topic_metadata => error_code topic is_internal [partition_metadata]
            error_code => INT16
            topic => STRING
            is_internal => BOOLEAN
            partition_metadata => error_code partition leader [replicas] [isr]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
    """
    def __init__(self, buff):
        fmt = 'i [iSiS] Si [hSb [hii [i] [i] ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        throttle_time_ms, broker_info, cluster_id, controller_id, topics = response
        self._populate(broker_info, topics, controller_id=controller_id,
                       cluster_id=cluster_id, throttle_time_ms=throttle_time_ms)


class MetadataResponseV4(MetadataResponseV3):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 4) => throttle_time_ms [brokers] cluster_id controller_id [topic_metadata]
        throttle_time_ms => INT32
        brokers => node_id host port rack
            node_id => INT32
            host => STRING
            port => INT32
            rack => NULLABLE_STRING
        cluster_id => NULLABLE_STRING
        controller_id => INT32
        topic_metadata => error_code topic is_internal [partition_metadata]
            error_code => INT16
            topic => STRING
            is_internal => BOOLEAN
            partition_metadata => error_code partition leader [replicas] [isr]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
    """
    pass


class MetadataResponseV5(MetadataResponseV4):
    """Response from MetadataRequest
    Specification::
    Metadata Response (Version: 5) => throttle_time_ms [brokers] cluster_id controller_id [topic_metadata]
        throttle_time_ms => INT32
        brokers => node_id host port rack
            node_id => INT32
            host => STRING
            port => INT32
            rack => NULLABLE_STRING
        cluster_id => NULLABLE_STRING
        controller_id => INT32
        topic_metadata => error_code topic is_internal [partition_metadata]
            error_code => INT16
            topic => STRING
            is_internal => BOOLEAN
            partition_metadata => error_code partition leader [replicas] [isr] [offline_replicas]
                error_code => INT16
                partition => INT32
                leader => INT32
                replicas => INT32
                isr => INT32
                offline_replicas => INT32  (new since v4)
    """
    def __init__(self, buff):
        fmt = 'i [iSiS] Si [hSb [hii [i] [i] [i]] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        throttle_time_ms, broker_info, cluster_id, controller_id, topics = response
        self._populate(broker_info, topics, controller_id=controller_id,
                       cluster_id=cluster_id, throttle_time_ms=throttle_time_ms)

    def _build_partition_metas(self, partitions):
        part_metas = {}
        for (p_err, id_, leader, replicas, isr, offline_replicas) in partitions:
            part_metas[id_] = PartitionMetadataV5(id_, leader, replicas,
                                                  isr, offline_replicas, p_err)
        return part_metas



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["recvall_into"]
from pykafka.exceptions import SocketDisconnectedError


def recvall_into(socket, bytea, size):
    """
    Reads `size` bytes from the socket into the provided bytearray (modifies
    in-place.)

    This is basically a hack around the fact that `socket.recv_into` doesn't
    allow buffer offsets.

    :type socket: :class:`socket.Socket`
    :type bytea: ``bytearray``
    :type size: int
    :rtype: `bytearray`
    """
    offset = 0
    while offset < size:
        remaining = size - offset
        try:
            chunk = socket.recv(remaining)
        except IOError:
            chunk = None
        if chunk is None or len(chunk) == 0:
            raise SocketDisconnectedError
        bytea[offset:(offset + len(chunk))] = chunk
        offset += len(chunk)
    return bytea



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["ResponseFuture", "Handler", "ThreadingHandler", "RequestHandler"]

from collections import namedtuple
import logging
import socket as pysocket
from socket import error as socket_error
from socket import gaierror as gaierror
import sys as _sys
import threading
import time

try:
    import gevent
    import gevent.event
    import gevent.lock
    import gevent.queue
    import gevent.socket as gsocket
    from gevent.socket import error as gsocket_error
    from gevent.socket import gaierror as g_gaierror
except ImportError:
    gevent = None

from .utils.compat import Queue, Empty, Semaphore

log = logging.getLogger(__name__)


class ResponseFuture(object):
    """A response which may have a value at some point."""

    def __init__(self, handler):
        """
        :type handler: :class:`pykafka.handlers.Handler`
        """
        self.handler = handler
        self.error = False
        self._ready = handler.Event()

    def set_response(self, response):
        """Set response data and trigger get method."""
        self.response = response
        self._ready.set()

    def set_error(self, error):
        """Set error and trigger get method."""
        self.error = error
        self._ready.set()

    def get(self, response_cls=None, timeout=None, **response_kwargs):
        """Block until data is ready and return.

        Raises an exception if there was an error.
        """
        self._ready.wait(timeout)
        if self.error:
            raise self.error
        if response_cls:
            return response_cls(self.response, **response_kwargs)
        else:
            return self.response


class Handler(object):
    """Base class for Handler classes"""
    def spawn(self, target, *args, **kwargs):
        """Create the worker that will process the work to be handled"""
        raise NotImplementedError


class ThreadingHandler(Handler):
    """A handler that uses a :class:`threading.Thread` to perform its work"""
    Queue = Queue
    Event = threading.Event
    Lock = threading.Lock
    Semaphore = Semaphore
    Socket = pysocket
    SockErr = socket_error
    GaiError = gaierror
    _workers_spawned = 0

    def sleep(self, seconds=0):
        time.sleep(seconds)

    # turn off RLock's super annoying default logging if possible
    def RLock(*args, **kwargs):
        kwargs['verbose'] = False
        try:
            return threading.RLock(*args[1:], **kwargs)
        except TypeError:
            kwargs.pop('verbose')
            return threading.RLock(*args[1:], **kwargs)

    def spawn(self, target, *args, **kwargs):
        if 'name' in kwargs:
            kwargs['name'] = "{}: {}".format(ThreadingHandler._workers_spawned, kwargs['name'])
        t = threading.Thread(target=target, *args, **kwargs)
        t.daemon = True
        t.start()
        ThreadingHandler._workers_spawned += 1
        return t


if gevent:
    class GEventHandler(Handler):
        """A handler that uses a greenlet to perform its work"""
        Queue = gevent.queue.JoinableQueue
        Event = gevent.event.Event
        Lock = gevent.lock.RLock  # fixme
        RLock = gevent.lock.RLock
        Semaphore = gevent.lock.Semaphore
        Socket = gsocket
        SockErr = gsocket_error
        GaiError = g_gaierror

        def sleep(self, seconds=0):
            gevent.sleep(seconds)

        def spawn(self, target, *args, **kwargs):
            # Greenlets don't support naming
            if 'name' in kwargs:
                kwargs.pop('name')
            t = gevent.spawn(target, *args, **kwargs)
            return t


class RequestHandler(object):
    """Uses a Handler instance to dispatch requests."""

    Task = namedtuple('Task', ['request', 'future'])
    Shared = namedtuple('Shared', ['connection', 'requests', 'ending'])

    def __init__(self, handler, connection):
        """
        :type handler: :class:`pykafka.handlers.Handler`
        :type connection: :class:`pykafka.connection.BrokerConnection`
        """
        self.handler = handler

        # NB self.shared is referenced directly by _start_thread(), so be careful not to
        # rebind it
        self.shared = self.Shared(connection=connection,
                                  requests=handler.Queue(),
                                  ending=handler.Event())

    def __del__(self):
        self.stop()

    def request(self, request, has_response=True):
        """Construct a new request

        :type request: :class:`pykafka.protocol.Request`
        :param has_response: Whether this request will return a response
        :returns: :class:`pykafka.handlers.ResponseFuture`
        """
        future = None
        if has_response:
            future = ResponseFuture(self.handler)

        task = self.Task(request, future)
        self.shared.requests.put(task)
        return future

    def start(self):
        """Start the request processor."""
        self.t = self._start_thread()

    def stop(self):
        """Stop the request processor."""
        shared = self.shared
        self.shared = None
        if log:
            log.info("RequestHandler.stop: about to flush requests queue")
        if shared:
            shared.requests.join()
            shared.ending.set()

    def _start_thread(self):
        """Run the request processor"""
        # We pass a direct reference to `shared` into the worker, to avoid
        # that thread holding a ref to `self`, which would prevent GC.  A
        # previous version of this used a weakref to `self`, but would
        # potentially abort the thread before the requests queue was empty
        shared = self.shared

        def worker():
            try:
                while not shared.ending.is_set():
                    try:
                        # set a timeout so we check `ending` every so often
                        task = shared.requests.get(timeout=1)
                    except Empty:
                        continue
                    try:
                        shared.connection.request(task.request)
                        if task.future:
                            res = shared.connection.response()
                            task.future.set_response(res)
                    except Exception as e:
                        if task.future:
                            task.future.set_error(e)
                    finally:
                        shared.requests.task_done()
                log.info("RequestHandler worker: exiting cleanly")
            except:
                # deal with interpreter shutdown in the same way that
                # python 3.x's threading module does, swallowing any
                # errors raised when core modules such as sys have
                # already been destroyed
                if _sys is None:
                    return
                raise

        name = "pykafka.RequestHandler.worker for {}:{}".format(
            self.shared.connection.host, self.shared.connection.port)
        return self.handler.spawn(worker, name=name)



# - coding: utf-8 -
"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Message", "CompressionType", "OffsetType"]
import datetime as dt
import logging


log = logging.getLogger(__name__)
EPOCH = dt.datetime(1970, 1, 1)


class Message(object):
    """Message class.

    :ivar response_code: Response code from Kafka
    :ivar topic: Originating topic
    :ivar payload: Message payload
    :ivar key: (optional) Message key
    :ivar offset: Message offset
    """
    __slots__ = []


class CompressionType(object):
    """Enum for the various compressions supported.

    :cvar NONE: Indicates no compression in use
    :cvar GZIP: Indicates gzip compression in use
    :cvar SNAPPY: Indicates snappy compression in use
    """
    NONE = 0
    GZIP = 1
    SNAPPY = 2
    LZ4 = 3


class OffsetType(object):
    """Enum for special values for earliest/latest offsets.

    :cvar EARLIEST: Indicates the earliest offset available for a partition
    :cvar LATEST: Indicates the latest offset available for a partition
    """
    EARLIEST = -2
    LATEST = -1



from contextlib import contextmanager
import unittest2

import pytest

from pykafka.exceptions import RdKafkaStoppedException, RdKafkaException
try:
    from pykafka.rdkafka import _rd_kafka
    RDKAFKA = True
except ImportError:
    RDKAFKA = False  # C extension not built
from pykafka.test.utils import get_cluster, stop_cluster
from pykafka.utils.compat import get_bytes


@pytest.mark.skipif(not RDKAFKA, reason="C extension for librdkafka not built.")
class TestRdKafkaConsumer(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = b'test-rdkafka-consumer'
        cls.n_partitions = 3
        cls.kafka.create_topic(cls.topic_name, cls.n_partitions, 2)
        cls.partition_ids = list(range(cls.n_partitions))
        cls.start_offsets = cls.n_partitions * [0]

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    @contextmanager
    def assert_thread_cnt_non_increasing(self):
        """Assert the wrapped function did not leave rdkafka threads behind

        Helper function to check if we correctly clean up rdkafka threads.  We
        can't just check that thread count dropped to zero, because finalisers
        on pypy may run much later than they do on cpython, so that threads
        spawned in previous tests may not have exited yet
        """
        preexisting_threads = _rd_kafka._thread_cnt()
        yield
        try:
            # We could just put `time.sleep(1)` here, but this lets the test
            # finish a bit faster - except on pypy, where it may time out when
            # preexisting_threads is nonzero
            _rd_kafka._wait_destroyed(1000)
        except:
            pass
        self.assertLessEqual(_rd_kafka._thread_cnt(), preexisting_threads)

    def test_start_fail(self):
        """See if Consumer_start cleans up upon failure"""
        with self.assert_thread_cnt_non_increasing():
            consumer = _rd_kafka.Consumer()
            with self.assertRaises(RdKafkaException):
                consumer.start(brokers=b"",  # this causes the exception
                               topic_name=self.topic_name,
                               partition_ids=self.partition_ids,
                               start_offsets=self.start_offsets)

    def test_stop(self):
        """Check Consumer_stop really shuts down the librdkafka consumer

        This is to deal with the fact that librdkafka's _destroy functions are
        all async, and therefore we don't get direct feedback if we didn't
        clean up in the correct order, yet the underlying consumer may remain
        up even if the python object is long gone.  Getting a zero thread
        count in the test gives some reassurance that we didn't leave any
        loose ends.
        """
        with self.assert_thread_cnt_non_increasing():
            consumer = _rd_kafka.Consumer()
            consumer.configure(conf=[])
            consumer.configure(topic_conf=[])
            consumer.start(brokers=get_bytes(self.kafka.brokers),
                           topic_name=self.topic_name,
                           partition_ids=self.partition_ids,
                           start_offsets=self.start_offsets)
            consumer.consume(100)  # just to reliably get some threads going
            consumer.stop()

    def test_stopped_exception(self):
        """Check Consumer_consume raises exception if handle was stopped"""
        consumer = _rd_kafka.Consumer(brokers=get_bytes(self.kafka.brokers),
                                      topic_name=self.topic_name,
                                      partition_ids=self.partition_ids,
                                      start_offsets=self.start_offsets)
        consumer.stop()
        with self.assertRaises(RdKafkaStoppedException):
            consumer.consume(1)



from contextlib import contextmanager
import logging
from pkg_resources import parse_version
import sys
import time

from pykafka.exceptions import RdKafkaStoppedException, ConsumerStoppedException
from pykafka.simpleconsumer import SimpleConsumer, OffsetType
from pykafka.utils.compat import get_bytes
from pykafka.utils.error_handlers import valid_int
from . import _rd_kafka
from . import helpers

log = logging.getLogger(__name__)


class RdKafkaSimpleConsumer(SimpleConsumer):
    """A pykafka.SimpleConsumer with librdkafka-based fetchers

    This aims to conform to the SimpleConsumer interface as closely as
    possible.  There are some notable differences:

    1. rotating over partitions: while message ordering within partitions is
       conserved (of course!), the order in which partitions are visited will
       deviate.  In particular, here we may emit more than one message from
       the same partition before visiting another
    2. ignores num_consumer_fetchers: librdkafka will typically spawn at least
       as many threads as there are kafka cluster nodes

    For an overview of how configuration keys are mapped to librdkafka's, see
    _mk_rdkafka_config_lists.

    The `broker_version` argument on `KafkaClient` must be set correctly to use the
    rdkafka consumer.
    """
    def __init__(self,
                 topic,
                 cluster,
                 consumer_group=None,
                 partitions=None,
                 fetch_message_max_bytes=1024 * 1024,
                 num_consumer_fetchers=1,
                 auto_commit_enable=False,
                 auto_commit_interval_ms=60 * 1000,
                 queued_max_messages=10**5,  # NB differs from SimpleConsumer
                 fetch_min_bytes=1,
                 fetch_error_backoff_ms=500,
                 fetch_wait_max_ms=100,
                 offsets_channel_backoff_ms=1000,
                 offsets_commit_max_retries=5,
                 auto_offset_reset=OffsetType.EARLIEST,
                 consumer_timeout_ms=-1,
                 auto_start=True,
                 reset_offset_on_start=False,
                 compacted_topic=False,
                 generation_id=-1,
                 consumer_id=b'',
                 deserializer=None,
                 reset_offset_on_fetch=True):
        callargs = {k: v for k, v in vars().items()
                         if k not in ("self", "__class__")}
        self._rdk_consumer = None
        self._poller_thread = None
        self._stop_poller_thread = cluster.handler.Event()
        self._broker_version = cluster._broker_version
        self._fetch_error_backoff_ms = valid_int(fetch_error_backoff_ms)
        # super() must come last for the case where auto_start=True
        super(RdKafkaSimpleConsumer, self).__init__(**callargs)

    def _setup_fetch_workers(self):
        brokers = b','.join(b.host + b":" + get_bytes(str(b.port))
                           for b in self._cluster.brokers.values())
        partition_ids = list(self._partitions_by_id.keys())
        start_offsets = [
            self._partitions_by_id[p].next_offset for p in partition_ids]
        conf, topic_conf = self._mk_rdkafka_config_lists()

        self._rdk_consumer = _rd_kafka.Consumer()
        log.debug("Configuring _rdk_consumer...")
        self._rdk_consumer.configure(conf=conf)
        self._rdk_consumer.configure(topic_conf=topic_conf)

        start_kwargs = {"brokers": brokers,
                        "topic_name": self._topic.name,
                        "partition_ids": partition_ids,
                        "start_offsets": start_offsets}
        log.debug("Starting _rdk_consumer with {}".format(start_kwargs))
        self._rdk_consumer.start(**start_kwargs)

        # Poll: for a consumer, the main reason to poll the handle is that
        # this de-queues log messages at error level that might otherwise be
        # held up in librdkafka
        def poll(rdk_handle, stop_event):
            while not stop_event.is_set():
                try:
                    rdk_handle.poll(timeout_ms=1000)
                except RdKafkaStoppedException:
                    break
                except:
                    self._worker_exception = sys.exc_info()
            log.debug("Exiting RdKafkaSimpleConsumer poller thread cleanly.")

        self._stop_poller_thread.clear()
        self._poller_thread = self._cluster.handler.spawn(
            poll, args=(self._rdk_consumer, self._stop_poller_thread))

    def consume(self, block=True, unblock_event=None):
        timeout_ms = self._consumer_timeout_ms if block else 1
        try:
            msg = self._consume(timeout_ms, unblock_event)
        # if _rdk_consumer is None we'll catch an AttributeError here
        except (RdKafkaStoppedException, AttributeError) as e:
            if not self._running:
                raise ConsumerStoppedException
            else:  # unexpected other reason
                raise
        else:
            if not self._running:
                # Even if we did get a msg back, we'll still want to abort
                # here, because the new offset wouldn't get committed anymore
                raise ConsumerStoppedException
        if msg is not None:
            # set offset in OwnedPartition so the autocommit_worker can find it
            self._partitions_by_id[msg.partition_id].set_offset(msg.offset)
        return msg

    def _consume(self, timeout_ms, unblock_event):
        """Helper to allow catching interrupts around rd_kafka_consume"""
        inner_timeout_ms = 500  # unblock at this interval at least

        if timeout_ms < 0:
            while True:
                self._raise_worker_exceptions()
                if unblock_event and unblock_event.is_set():
                    return
                msg = self._rdk_consumer.consume(inner_timeout_ms)
                if msg is not None:
                    return msg
        else:
            t_start = time.time()
            leftover_ms = timeout_ms
            while leftover_ms > 0:
                self._raise_worker_exceptions()
                if unblock_event and unblock_event.is_set():
                    return
                inner_timeout_ms = int(min(leftover_ms, inner_timeout_ms))
                msg = self._rdk_consumer.consume(inner_timeout_ms)
                if msg is not None:
                    return msg
                elapsed_ms = 1000 * (time.time() - t_start)
                leftover_ms = timeout_ms - elapsed_ms

    def stop(self):
        # NB we should always call super() first, because it takes care of
        # shipping all important state (ie stored offsets) out
        ret = super(RdKafkaSimpleConsumer, self).stop()
        if self._rdk_consumer is not None:
            self._stop_poller_thread.set()
            # Call _rd_kafka.Consumer.stop explicitly, so we may catch errors:
            self._rdk_consumer.stop()
            log.debug("Issued stop to _rdk_consumer.")
            self._rdk_consumer = None
        return ret

    def fetch_offsets(self):
        # Restart, because _rdk_consumer needs its internal offsets resynced
        with self._stop_start_rdk_consumer():
            return super(RdKafkaSimpleConsumer, self).fetch_offsets()

    def reset_offsets(self, partition_offsets=None):
        # Restart, because _rdk_consumer needs its internal offsets resynced
        with self._stop_start_rdk_consumer():
            return super(
                RdKafkaSimpleConsumer, self).reset_offsets(partition_offsets)

    @contextmanager
    def _stop_start_rdk_consumer(self):
        """Context manager for methods to temporarily stop _rdk_consumer

        We need this because we hold read-offsets both in pykafka and
        internally in _rdk_consumer.  We'll hold the one in pykafka to be the
        ultimate source of truth.  So whenever offsets are to be changed (other
        than through consume() that is), we need to clobber _rdk_consumer.
        """
        restart_required = self._running and self._rdk_consumer is not None
        if restart_required:
            # Note we must not call a full self.stop() as that would stop
            # SimpleConsumer threads too, and if we'd have to start() again
            # that could have other side effects (eg resetting offsets).
            self._rdk_consumer.stop()
            log.debug("Temporarily stopped _rdk_consumer.")
        yield
        if restart_required:
            self._setup_fetch_workers()
            log.debug("Restarted _rdk_consumer.")

    def _mk_rdkafka_config_lists(self):
        """Populate conf, topic_conf to configure the rdkafka consumer"""
        # For documentation purposes, all consumer-relevant settings (all those
        # marked 'C' or '*') that appear in librdkafka/CONFIGURATION.md should
        # be listed below, in either `conf` or `topic_conf`, even if we do not
        # set them and they are commented out.
        ver10 = parse_version(self._broker_version) >= parse_version("0.10.0")
        conf = {  # destination: rd_kafka_conf_set
            "client.id": "pykafka.rdkafka",
            # Handled via rd_kafka_brokers_add instead:
            ##"metadata.broker.list"

            # NB these refer not to payloads, but to wire messages
            ##"message.max.bytes"  # leave at default
            "receive.message.max.bytes": (  # ~ sum of PartitionFetchRequests
                self._fetch_message_max_bytes * (len(self.partitions) + 1)),

            # No direct equivalents:
            ##"metadata.request.timeout.ms"
            ##"topic.metadata.refresh.interval.ms"
            ##"topic.metadata.refresh.fast.cnt"
            ##"topic.metadata.refresh.fast.interval.ms"
            ##"topic.metadata.refresh.sparse"

            ##"debug": "all",

            "socket.timeout.ms": self._cluster._socket_timeout_ms,
            ##"socket.send.buffer.bytes"
            ##"socket.receive.buffer.bytes"
            ##"socket.keepalive.enable"
            ##"socket.max.fails"
            ##"broker.address.ttl"
            ##"broker.address.family"

            # None of these need to be hooked up
            ##"statistics.interval.ms"
            ##"error_cb"  # we let errors be reported via log_cb
            ##"stats_cb"

            ##"log_cb"  # gets set in _rd_kafka module
            ##"log_level": 7,

            ##"socket_cb"
            ##"open_cb"
            ##"opaque"
            ##"internal.termination.signal"

            # Although the names seem to disagree, SimpleConsumer currently
            # uses _queued_max_messages in a way analogous to
            # queued.min.messages; that is "keep trying to fetch until there's
            # this number of messages on the queue".  There's no equivalent of
            # queued.max.messages.kbytes so for now we infer the implied
            # maximum (which, with default settings, is ~2GB per partition):
            "queued.min.messages": self._queued_max_messages,
            "queued.max.messages.kbytes": str(
                self._queued_max_messages
                * self._fetch_message_max_bytes // 1024),

            "fetch.wait.max.ms": self._fetch_wait_max_ms,
            "fetch.message.max.bytes": self._fetch_message_max_bytes,
            "fetch.min.bytes": self._fetch_min_bytes,
            "fetch.error.backoff.ms": self._fetch_error_backoff_ms,
            "api.version.request": ver10,

            # We're outsourcing message fetching, but not offset management or
            # consumer rebalancing to librdkafka.  Thus, consumer-group id
            # *must* remain unset, or we would appear to be two consumer
            # instances to the kafka cluster:
            ##"group.id"
            }
        # broker.version.fallback is incompatible with >-0.10
        if not ver10:
            conf["broker.version.fallback"] = self._broker_version
        conf.update(helpers.rdk_ssl_config(self._cluster))

        map_offset_types = {
            OffsetType.EARLIEST: "smallest",
            OffsetType.LATEST: "largest",
            }
        topic_conf = {
            ##"opaque"
            ##"group.id"  # see note above re group.id

            # pykafka handles offset commits
            "auto.commit.enable": "false",
            ##"auto.commit.interval.ms"
            "auto.offset.reset": map_offset_types[self._auto_offset_reset],
            ##"offset.store.path"
            ##"offset.store.sync.interval.ms"
            ##"offset.store.method"
            }
        # librdkafka expects all config values as strings:
        conf = [(key, str(conf[key])) for key in conf]
        topic_conf = [(key, str(topic_conf[key])) for key in topic_conf]
        return conf, topic_conf



# - coding: utf-8 -
from datetime import datetime
import struct
from pkg_resources import parse_version
from six import integer_types
from zlib import crc32

from ..common import CompressionType, Message
from ..exceptions import MessageSetDecodeFailure
from ..utils import Serializable, struct_helpers, compression
from ..utils.compat import buffer


class Message(Message, Serializable):
    """Representation of a Kafka Message

    NOTE: Compression is handled in the protocol because of the way Kafka embeds
    compressed MessageSets within Messages

    Specification::

        Message => Crc MagicByte Attributes Key Value
          Crc => int32
          MagicByte => int8
          Attributes => int8
          Key => bytes
          Value => bytes

    :class:`pykafka.protocol.Message` also contains `partition` and
    `partition_id` fields. Both of these have meaningless default values. When
    :class:`pykafka.protocol.Message` is used by the producer, `partition_id`
    identifies the Message's destination partition.
    When used in a :class:`pykafka.protocol.FetchRequest`, `partition_id`
    is set to the id of the partition from which the message was sent on
    receipt of the message. In the :class:`pykafka.simpleconsumer.SimpleConsumer`,
    `partition` is set to the :class:`pykafka.partition.Partition` instance
    from which the message was sent.

    :ivar compression_type: The compression algorithm used to generate the message's
        current value. Internal use only - regardless of the algorithm used, this
        will be `CompressionType.NONE` in any publicly accessible `Message`s.
    :ivar partition_key: Value used to assign this message to a particular partition.
    :ivar value: The payload associated with this message
    :ivar offset: The offset of the message
    :ivar partition_id: The id of the partition to which this message belongs
    :ivar delivery_report_q: For use by :class:`pykafka.producer.Producer`
    """

    __slots__ = [
        "compression_type",
        "partition_key",
        "value",
        "offset",
        "partition_id",
        "partition",
        "produce_attempt",
        "delivery_report_q",
        "protocol_version",
        "timestamp"
    ]
    VALID_TS_TYPES = integer_types + (float, type(None))

    def __init__(self,
                 value,
                 partition_key=None,
                 compression_type=CompressionType.NONE,
                 offset=-1,
                 partition_id=-1,
                 produce_attempt=0,
                 protocol_version=0,
                 timestamp=None,
                 delivery_report_q=None):
        self.compression_type = compression_type
        self.partition_key = partition_key
        self.value = value
        self.offset = offset
        if timestamp is None and protocol_version > 0:
            timestamp = datetime.utcnow()
        self.set_timestamp(timestamp)
        # this is set on decode to expose it to clients that use the protocol
        # implementation but not the consumer
        self.partition_id = partition_id
        # self.partition is set by the consumer
        self.partition = None
        self.produce_attempt = produce_attempt
        # delivery_report_q is used by the producer
        self.delivery_report_q = delivery_report_q
        assert protocol_version in (0, 1)
        self.protocol_version = protocol_version

    def __len__(self):
        size = 4 + 1 + 1 + 4 + 4
        if self.value is not None:
            size += len(self.value)
        if self.partition_key is not None:
            size += len(self.partition_key)
        if self.protocol_version > 0 and self.timestamp:
            size += 8
        return size

    @classmethod
    def decode(self, buff, msg_offset=-1, partition_id=-1):
        (crc, protocol_version, attr) = struct_helpers.unpack_from('iBB', buff, 0)
        offset = 6
        timestamp = 0
        if protocol_version > 0:
            (timestamp,) = struct_helpers.unpack_from('Q', buff, offset)
            offset += 8
        (key, val) = struct_helpers.unpack_from('YY', buff, offset)
        # TODO: Handle CRC failure
        return Message(val,
                       partition_key=key,
                       compression_type=attr,
                       offset=msg_offset,
                       protocol_version=protocol_version,
                       timestamp=timestamp,
                       partition_id=partition_id)

    def pack_into(self, buff, offset):
        """Serialize and write to ``buff`` starting at offset ``offset``.

        Intentionally follows the pattern of ``struct.pack_into``

        :param buff: The buffer to write into
        :param offset: The offset to start the write at
        """
        # NB a length of 0 means an empty string, whereas -1 means null
        # Assuming a CreateTime timestamp, not a LogAppendTime.
        len_key = -1 if self.partition_key is None else len(self.partition_key)
        len_value = -1 if self.value is None else len(self.value)
        protocol_version = self.protocol_version
        # Only actually use protocol 1 if timestamp is defined.
        if self.protocol_version == 1 and self.timestamp:
            fmt = '!BBQi%dsi%ds' % (max(len_key, 0), max(len_value, 0))
        else:
            protocol_version = 0
            fmt = '!BBi%dsi%ds' % (max(len_key, 0), max(len_value, 0))
        args = [protocol_version,
                self.compression_type,
                len_key,
                self.partition_key or b"",
                len_value,
                self.value or b""]
        if protocol_version > 0:
            args.insert(2, int(self.timestamp))
        struct.pack_into(fmt, buff, offset + 4, *args)
        fmt_size = struct.calcsize(fmt)
        data = buffer(buff[(offset + 4):(offset + 4 + fmt_size)])
        crc = crc32(data) & 0xffffffff
        struct.pack_into('!I', buff, offset, crc)

    @property
    def timestamp_dt(self):
        """Get the timestamp as a datetime, if valid"""
        if self.timestamp > 0:
            # Assuming a unix epoch
            return datetime.utcfromtimestamp(self.timestamp / 1000.0)

    @timestamp_dt.setter
    def timestamp_dt(self, dt):
        """Set the timestamp from a datetime object"""
        self.timestamp = int(
            1000 * (dt - datetime(1970, 1, 1)).total_seconds())

    def set_timestamp(self, ts):
        if type(ts) in self.VALID_TS_TYPES:
            self.timestamp = ts
        elif type(ts) == datetime:
            self.timestamp_dt = ts
        else:
            raise RuntimeError()


class MessageSet(Serializable):
    """Representation of a set of messages in Kafka

    This isn't useful outside of direct communications with Kafka, so we
    keep it hidden away here.

    N.B.: MessageSets are not preceded by an int32 like other array elements in the
    protocol.

    Specification::

        MessageSet => [Offset MessageSize Message]
          Offset => int64
          MessageSize => int32
    """
    def __init__(self,
                 compression_type=CompressionType.NONE,
                 messages=None,
                 broker_version='0.9.0'):
        """Create a new MessageSet

        :param compression_type: Compression to use on the messages
        :param messages: An initial list of messages for the set
        :param broker_version: A broker version with which this MessageSet is compatible
        """
        self.compression_type = compression_type
        self._messages = messages or []
        self._compressed = None  # compressed Message if using compression
        self._broker_version = broker_version

    def __len__(self):
        """Length of the serialized message, in bytes

        We don't put the MessageSetSize in front of the serialization
        because that's *technically* not part of the MessageSet. Most
        requests/responses using MessageSets need that size, though, so
        be careful when using this.
        """
        if self.compression_type == CompressionType.NONE:
            messages = self._messages
        else:
            # The only way to get __len__ of compressed is to compress.
            # Store that so we don't have to do it twice
            if self._compressed is None:
                self._compressed = self._get_compressed()
            messages = [self._compressed]
        return (8 + 4) * len(messages) + sum(len(m) for m in messages)

    @property
    def messages(self):
        # Make sure accessing messages directly clears cached compressed data
        self._compressed = None
        return self._messages

    def _get_compressed(self):
        """Get a compressed representation of all current messages.

        Returns a Message object with correct headers set and compressed
        data in the value field.
        """
        assert self.compression_type != CompressionType.NONE
        tmp_mset = MessageSet(messages=self._messages)
        uncompressed = bytearray(len(tmp_mset))
        tmp_mset.pack_into(uncompressed, 0)
        if self.compression_type == CompressionType.GZIP:
            compressed = compression.encode_gzip(buffer(uncompressed))
        elif self.compression_type == CompressionType.SNAPPY:
            compressed = compression.encode_snappy(buffer(uncompressed))
        elif self.compression_type == CompressionType.LZ4:
            if parse_version(self._broker_version) >= parse_version('0.10.0'):
                compressed = compression.encode_lz4(buffer(uncompressed))
            else:
                compressed = compression.encode_lz4_old_kafka(buffer(uncompressed))
        else:
            raise TypeError("Unknown compression: %s" % self.compression_type)
        protocol_version = max((m.protocol_version for m in self._messages))
        return Message(compressed, compression_type=self.compression_type,
                       protocol_version=protocol_version)

    @classmethod
    def decode(cls, buff, partition_id=-1):
        """Decode a serialized MessageSet."""
        messages = []
        offset = 0
        attempted = False
        while offset < len(buff):
            # if the buffer is not large enough to contain the offset information
            if len(buff) - offset < 12:
                break
            msg_offset, size = struct.unpack_from('!qi', buff, offset)
            offset += 12
            attempted = True
            # if the buffer is not large enough to contain the full message
            if len(buff) - offset < size:
                break
            # TODO: Check we have all the requisite bytes
            message = Message.decode(buff[offset:offset + size],
                                     msg_offset,
                                     partition_id=partition_id)
            messages.append(message)
            offset += size
        if len(messages) == 0 and attempted:
            raise MessageSetDecodeFailure(size)
        return MessageSet(messages=messages)

    def pack_into(self, buff, offset):
        """Serialize and write to ``buff`` starting at offset ``offset``.

        Intentionally follows the pattern of ``struct.pack_into``

        :param buff: The buffer to write into
        :param offset: The offset to start the write at
        """
        if self.compression_type == CompressionType.NONE:
            messages = self._messages
        else:
            if self._compressed is None:
                self._compressed = self._get_compressed()
            messages = [self._compressed]

        for message in messages:
            mlen = len(message)
            struct.pack_into('!qi', buff, offset, -1, mlen)
            offset += 12
            message.pack_into(buff, offset)
            offset += mlen



import lz4f
import platform
import pytest
import unittest2
from uuid import uuid4

from pykafka.utils import compression


class CompressionTests(unittest2.TestCase):
    """Keeping these simple by verifying what goes in is what comes out."""
    text = b"The man in black fled across the desert, and the gunslinger followed."

    def test_gzip(self):
        encoded = compression.encode_gzip(self.text)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_gzip(encoded)
        self.assertEqual(self.text, decoded)

    def test_snappy(self):
        encoded = compression.encode_snappy(self.text)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_snappy(encoded)
        self.assertEqual(self.text, decoded)

    def test_snappy_xerial(self):
        encoded = compression.encode_snappy(self.text, xerial_compatible=True)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_snappy(encoded)
        self.assertEqual(self.text, decoded)

    def test_snappy_large_payload(self):
        if platform.python_implementation() == "PyPy":
            pytest.skip("PyPy fails to compress large messages with Snappy")
        payload = b''.join([uuid4().bytes for i in range(10)])
        c = compression.encode_snappy(payload)
        self.assertEqual(compression.decode_snappy(c), payload)

    def test_lz4(self):
        encoded = compression.encode_lz4(self.text)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_lz4(encoded)
        self.assertEqual(self.text, decoded)

    def test_lz4f(self):
        if platform.python_implementation() == "PyPy":
            pytest.skip("lz4f is currently unsupported with PyPy")
        encoded = lz4f.compressFrame(self.text)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_lz4f(encoded)
        self.assertEqual(self.text, decoded)

    def test_lz4_old_kafka(self):
        encoded = compression.encode_lz4_old_kafka(self.text)
        self.assertNotEqual(self.text, encoded)

        decoded = compression.decode_lz4_old_kafka(encoded)
        self.assertEqual(self.text, decoded)


if __name__ == '__main__':
    unittest2.main()



from __future__ import division
"""
Author: Emmett Butler
"""
__license__ = """
Copyright 2016 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["ManagedBalancedConsumer"]
import logging
import sys
import uuid
import weakref

from .balancedconsumer import BalancedConsumer
from .common import OffsetType
from .exceptions import (IllegalGeneration, RebalanceInProgress, NotCoordinatorForGroup,
                         GroupCoordinatorNotAvailable, ERROR_CODES, GroupLoadInProgress,
                         UnicodeException)
from .membershipprotocol import RangeProtocol
from .protocol import MemberAssignment
from .utils.compat import iterkeys, get_string
from .utils.error_handlers import valid_int

log = logging.getLogger(__name__)


class ManagedBalancedConsumer(BalancedConsumer):
    """A self-balancing consumer that uses Kafka 0.9's Group Membership API

    Implements the Group Management API semantics for Kafka 0.9 compatibility

    Maintains a single instance of SimpleConsumer, periodically using the
    consumer rebalancing algorithm to reassign partitions to this
    SimpleConsumer.

    This class overrides the functionality of
    :class:`pykafka.balancedconsumer.BalancedConsumer` that deals with ZooKeeper and
    inherits other functionality directly.
    """
    def __init__(self,
                 topic,
                 cluster,
                 consumer_group,
                 fetch_message_max_bytes=1024 * 1024,
                 num_consumer_fetchers=1,
                 auto_commit_enable=False,
                 auto_commit_interval_ms=60 * 1000,
                 queued_max_messages=2000,
                 fetch_min_bytes=1,
                 fetch_error_backoff_ms=500,
                 fetch_wait_max_ms=100,
                 offsets_channel_backoff_ms=1000,
                 offsets_commit_max_retries=5,
                 auto_offset_reset=OffsetType.EARLIEST,
                 consumer_timeout_ms=-1,
                 rebalance_max_retries=5,
                 rebalance_backoff_ms=2 * 1000,
                 auto_start=True,
                 reset_offset_on_start=False,
                 post_rebalance_callback=None,
                 use_rdkafka=False,
                 compacted_topic=True,
                 heartbeat_interval_ms=3000,
                 membership_protocol=RangeProtocol,
                 deserializer=None,
                 reset_offset_on_fetch=True):
        """Create a ManagedBalancedConsumer instance

        :param topic: The topic this consumer should consume
        :type topic: :class:`pykafka.topic.Topic`
        :param cluster: The cluster to which this consumer should connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param consumer_group: The name of the consumer group this consumer
            should join. Consumer group names are namespaced at the cluster level,
            meaning that two consumers consuming different topics with the same group name
            will be treated as part of the same group.
        :type consumer_group: str
        :param fetch_message_max_bytes: The number of bytes of messages to
            attempt to fetch with each fetch request
        :type fetch_message_max_bytes: int
        :param num_consumer_fetchers: The number of workers used to make
            FetchRequests
        :type num_consumer_fetchers: int
        :param auto_commit_enable: If true, periodically commit to kafka the
            offset of messages already fetched by this consumer. This also
            requires that `consumer_group` is not `None`.
        :type auto_commit_enable: bool
        :param auto_commit_interval_ms: The frequency (in milliseconds) at which
            the consumer's offsets are committed to kafka. This setting is
            ignored if `auto_commit_enable` is `False`.
        :type auto_commit_interval_ms: int
        :param queued_max_messages: The maximum number of messages buffered for
            consumption in the internal
            :class:`pykafka.simpleconsumer.SimpleConsumer`
        :type queued_max_messages: int
        :param fetch_min_bytes: The minimum amount of data (in bytes) that the
            server should return for a fetch request. If insufficient data is
            available, the request will block until sufficient data is available.
        :type fetch_min_bytes: int
        :param fetch_error_backoff_ms: *UNUSED*.
            See :class:`pykafka.simpleconsumer.SimpleConsumer`.
        :type fetch_error_backoff_ms: int
        :param fetch_wait_max_ms: The maximum amount of time (in milliseconds)
            that the server will block before answering a fetch request if
            there isn't sufficient data to immediately satisfy `fetch_min_bytes`.
        :type fetch_wait_max_ms: int
        :param offsets_channel_backoff_ms: Backoff time to retry failed offset
            commits and fetches.
        :type offsets_channel_backoff_ms: int
        :param offsets_commit_max_retries: The number of times the offset commit
            worker should retry before raising an error.
        :type offsets_commit_max_retries: int
        :param auto_offset_reset: What to do if an offset is out of range. This
            setting indicates how to reset the consumer's internal offset
            counter when an `OffsetOutOfRangeError` is encountered.
        :type auto_offset_reset: :class:`pykafka.common.OffsetType`
        :param consumer_timeout_ms: Amount of time (in milliseconds) the
            consumer may spend without messages available for consumption
            before returning None.
        :type consumer_timeout_ms: int
        :param rebalance_max_retries: The number of times the rebalance should
            retry before raising an error.
        :type rebalance_max_retries: int
        :param rebalance_backoff_ms: Backoff time (in milliseconds) between
            retries during rebalance.
        :type rebalance_backoff_ms: int
        :param auto_start: Whether the consumer should start after __init__ is complete.
            If false, it can be started with `start()`.
        :type auto_start: bool
        :param reset_offset_on_start: Whether the consumer should reset its
            internal offset counter to `self._auto_offset_reset` and commit that
            offset immediately upon starting up
        :type reset_offset_on_start: bool
        :param post_rebalance_callback: A function to be called when a rebalance is
            in progress. This function should accept three arguments: the
            :class:`pykafka.balancedconsumer.BalancedConsumer` instance that just
            completed its rebalance, a dict of partitions that it owned before the
            rebalance, and a dict of partitions it owns after the rebalance. These dicts
            map partition ids to the most recently known offsets for those partitions.
            This function can optionally return a dictionary mapping partition ids to
            offsets. If it does, the consumer will reset its offsets to the supplied
            values before continuing consumption.
            Note that the BalancedConsumer is in a poorly defined state at
            the time this callback runs, so that accessing its properties
            (such as `held_offsets` or `partitions`) might yield confusing
            results.  Instead, the callback should really rely on the
            provided partition-id dicts, which are well-defined.
        :type post_rebalance_callback: function
        :param use_rdkafka: Use librdkafka-backed consumer if available
        :type use_rdkafka: bool
        :param compacted_topic: Set to read from a compacted topic. Forces
            consumer to use less stringent message ordering logic because compacted
            topics do not provide offsets in strict incrementing order.
        :type compacted_topic: bool
        :param heartbeat_interval_ms: The amount of time in milliseconds to wait between
            heartbeat requests
        :type heartbeat_interval_ms: int
        :param membership_protocol: The group membership protocol to which this consumer
            should adhere
        :type membership_protocol: :class:`pykafka.membershipprotocol.GroupMembershipProtocol`
        :param deserializer: A function defining how to deserialize messages returned
            from Kafka. A function with the signature d(value, partition_key) that
            returns a tuple of (deserialized_value, deserialized_partition_key). The
            arguments passed to this function are the bytes representations of a
            message's value and partition key, and the returned data should be these
            fields transformed according to the client code's serialization logic.
            See `pykafka.utils.__init__` for stock implemtations.
        :type deserializer: function
        :param reset_offset_on_fetch: Whether to update offsets during fetch_offsets.
               Disable for read-only use cases to prevent side-effects.
        :type reset_offset_on_fetch: bool
        """

        self._cluster = cluster
        try:
            self._consumer_group = get_string(consumer_group).encode('ascii')
        except UnicodeEncodeError:
            raise UnicodeException("Consumer group name '{}' contains non-ascii "
                                   "characters".format(consumer_group))
        self._topic = topic

        self._auto_commit_enable = auto_commit_enable
        self._auto_commit_interval_ms = valid_int(auto_commit_interval_ms)
        self._fetch_message_max_bytes = valid_int(fetch_message_max_bytes)
        self._fetch_min_bytes = valid_int(fetch_min_bytes)
        self._num_consumer_fetchers = valid_int(num_consumer_fetchers)
        self._queued_max_messages = valid_int(queued_max_messages)
        self._fetch_wait_max_ms = valid_int(fetch_wait_max_ms, allow_zero=True)
        self._consumer_timeout_ms = valid_int(consumer_timeout_ms,
                                              allow_zero=True, allow_negative=True)
        self._offsets_channel_backoff_ms = valid_int(offsets_channel_backoff_ms)
        self._offsets_commit_max_retries = valid_int(offsets_commit_max_retries,
                                                     allow_zero=True)
        self._auto_offset_reset = auto_offset_reset
        self._reset_offset_on_start = reset_offset_on_start
        self._rebalance_max_retries = valid_int(rebalance_max_retries, allow_zero=True)
        self._rebalance_backoff_ms = valid_int(rebalance_backoff_ms)
        self._post_rebalance_callback = post_rebalance_callback
        self._is_compacted_topic = compacted_topic
        self._membership_protocol = membership_protocol
        self._membership_protocol.metadata.topic_names = [self._topic.name]
        self._heartbeat_interval_ms = valid_int(heartbeat_interval_ms)
        self._deserializer = deserializer
        self._reset_offset_on_fetch = reset_offset_on_fetch
        if use_rdkafka is True:
            raise ImportError("use_rdkafka is not available for {}".format(
                self.__class__.__name__))
        self._use_rdkafka = use_rdkafka

        self._generation_id = -1
        self._rebalancing_lock = cluster.handler.Lock()
        self._rebalancing_in_progress = self._cluster.handler.Event()
        self._internal_consumer_running = self._cluster.handler.Event()
        # ManagedBalancedConsumers in the same process cannot share connections.
        # This connection hash is passed to Broker calls that use the group
        # membership  API
        self._connection_id = uuid.uuid4()
        self._consumer = None
        self._group_coordinator = None
        self._consumer_id = b''
        self._worker_exception = None
        self._default_error_handlers = self._build_default_error_handlers()

        if auto_start is True:
            self.start()

    def _setup_heartbeat_worker(self):
        """Start the heartbeat worker"""
        self = weakref.proxy(self)

        def fetcher():
            while True:
                try:
                    if not self._running:
                        break

                    log.info("Sending heartbeat from consumer '%s'", self._consumer_id)
                    res = self._group_coordinator.heartbeat(self._connection_id,
                                                            self._consumer_group,
                                                            self._generation_id,
                                                            self._consumer_id)
                    if res.error_code != 0:
                        log.info("Error code %d encountered on heartbeat.",
                                 res.error_code)
                        self._handle_error(res.error_code)
                        self._rebalance()

                    self._cluster.handler.sleep(self._heartbeat_interval_ms / 1000)
                except ReferenceError:
                    break
                except Exception:
                    # surface all exceptions to the main thread
                    self._worker_exception = sys.exc_info()
                    break
            log.debug("Heartbeat worker exiting")
        log.info("Starting heartbeat worker")
        return self._cluster.handler.spawn(
            fetcher, name="pykafka.ManagedBalancedConsumer.heartbeats")

    def start(self):
        """Start this consumer.

        Must be called before consume() if `auto_start=False`.
        """
        try:
            self._running = True
            self._group_coordinator = self._cluster.get_group_coordinator(
                self._consumer_group)
            self._rebalance()
            self._setup_heartbeat_worker()
        except Exception:
            log.exception("Stopping consumer in response to error")
            self.stop()

    def stop(self):
        """Stop this consumer

        Should be called as part of a graceful shutdown
        """
        self._running = False
        if self._consumer is not None:
            self._consumer.stop()
        if self._group_coordinator is not None:
            self._group_coordinator.leave_group(self._connection_id,
                                                self._consumer_group,
                                                self._consumer_id)

    def _update_member_assignment(self):
        """Join a managed consumer group and start consuming assigned partitions

        Equivalent to
        `pykafka.balancedconsumer.BalancedConsumer._update_member_assignment`,
        but uses the Kafka 0.9 Group Membership API instead of ZooKeeper to manage
        group state
        """
        for i in range(self._rebalance_max_retries):
            try:
                members = self._join_group()
                # generate partition assignments for each group member
                # if this is not the leader, join_result.members will be empty
                group_assignments = [
                    (member_id,
                     MemberAssignment([
                        (self._topic.name,
                         [p.id for p in self._membership_protocol.decide_partitions(
                          iterkeys(members), self._topic.partitions, member_id)])])
                    ) for member_id in members]

                assignment = self._sync_group(group_assignments)
                self._setup_internal_consumer(
                    partitions=[self._topic.partitions[pid] for pid in assignment[0][1]])
                log.debug("Successfully rebalanced consumer '%s'", self._consumer_id)
                break
            except Exception as ex:
                if i == self._rebalance_max_retries - 1:
                    log.warning('Failed to rebalance s after %d retries.', i)
                    raise
                log.exception(ex)
                log.info('Unable to complete rebalancing. Retrying')
                self._cluster.handler.sleep(i * (self._rebalance_backoff_ms / 1000))
        self._raise_worker_exceptions()

    def _build_default_error_handlers(self):
        """Set up default responses to common error codes"""
        self = weakref.proxy(self)

        def _handle_GroupCoordinatorNotAvailable():
            self._group_coordinator = self._cluster.get_group_coordinator(
                self._consumer_group)

        def _handle_NotCoordinatorForGroup():
            self._group_coordinator = self._cluster.get_group_coordinator(
                self._consumer_group)

        return {
            GroupCoordinatorNotAvailable.ERROR_CODE: _handle_GroupCoordinatorNotAvailable,
            NotCoordinatorForGroup.ERROR_CODE: _handle_NotCoordinatorForGroup,
            GroupLoadInProgress.ERROR_CODE: None,
            RebalanceInProgress.ERROR_CODE: None,
            IllegalGeneration.ERROR_CODE: None
        }

    def _handle_error(self, error_code):
        """Call the appropriate handler function for the given error code

        :param error_code: The error code returned from a Group Membership API request
        :type error_code: int
        """
        if error_code not in self._default_error_handlers:
            raise ERROR_CODES[error_code]()
        if self._default_error_handlers[error_code] is not None:
            self._default_error_handlers[error_code]()

    def _join_group(self):
        """Send a JoinGroupRequest.

        Assigns a member id and tells the coordinator about this consumer.
        """
        log.info("Sending JoinGroupRequest for consumer id '%s'", self._consumer_id)
        for i in range(self._cluster._max_connection_retries):
            join_result = self._group_coordinator.join_group(self._connection_id,
                                                             self._consumer_group,
                                                             self._consumer_id,
                                                             self._topic.name,
                                                             self._membership_protocol)
            if join_result.error_code == 0:
                break
            log.info("Error code %d encountered during JoinGroupRequest for"
                     " generation '%s'", join_result.error_code, self._generation_id)
            if i == self._cluster._max_connection_retries - 1:
                raise ERROR_CODES[join_result.error_code]
            self._handle_error(join_result.error_code)
            self._cluster.handler.sleep(i * 2)
        self._generation_id = join_result.generation_id
        self._consumer_id = join_result.member_id
        return join_result.members

    def _sync_group(self, group_assignments):
        """Send a SyncGroupRequest.

        If this consumer is the group leader, this call informs the other consumers
        of their partition assignments. For all consumers including the leader, this call
        is used to fetch partition assignments.

        The group leader *could* tell itself its own assignment instead of using the
        result of this request, but it does the latter to ensure consistency.
        """
        log.info("Sending SyncGroupRequest for consumer id '%s'", self._consumer_id)
        for i in range(self._cluster._max_connection_retries):
            sync_result = self._group_coordinator.sync_group(self._connection_id,
                                                             self._consumer_group,
                                                             self._generation_id,
                                                             self._consumer_id,
                                                             group_assignments)
            if sync_result.error_code == 0:
                break
            log.info("Error code %d encountered during SyncGroupRequest",
                     sync_result.error_code)
            if i == self._cluster._max_connection_retries - 1:
                raise ERROR_CODES[sync_result.error_code]
            self._handle_error(sync_result.error_code)
            self._cluster.handler.sleep(i * 2)
        return sync_result.member_assignment.partition_assignment



import platform

import pytest

from tests.pykafka import test_ssl


@pytest.mark.skipif(platform.python_implementation() == "PyPy",
                    reason="We pass PyObject pointers as msg_opaques for "
                           "delivery callbacks, which is unsafe on PyPy.")
class TestRdKafkaSsl(test_ssl.SslIntegrationTests):
    USE_RDKAFKA = True
